#!/usr/bin/perl

$| = 1;

# ddrwipe.pl   ("generate script to zero all the GOOD data sectors")
#
# This is a Perl program which reads a GNU ddrescue 1.x "log files" (really,
# sector list files), and generates a script to zero all the good data
# sectors, while leaving the bad ones along.  The purpose is to prepare a
# bad hard disk drive for warranty replacement.  (We don't want to zero the
# bad sectors, because the drive would probably "spare them out" so that
# the drive no longer appeared to be bad.)


# This program is UNCOPYRIGHTED.  It was written by:
#
#    Dave Burton
#    Burton Systems Software
#    Tel: +1 919-481-0149
#    Email: dave at burtonsys.com
#
#
# v.1 - 8-Aug-2006 - first version.  -DAB

#  =>keyflag<=-- "%v, %d"
$ver_ordinal_and_date = "5, 05-Oct-09";
# the TLIB Version Control version number and last-modified date


if (!defined $verbose) {
   $verbose = 0;
}
if (!defined $use_dd) {
   $use_dd = 0;
}
if (!defined $immediatew) {
   $immediatew = 0;
}
if (!defined $quick) {
   $quick = 0;
}

$perlver = "3 or earlier";
if ($] =~ /\$\$Revision\:\s*([0-9.]+)\s/) {
   $perlver = $1;  # probably 4.something
} elsif ($] =~ /([0-9][0-9.]*)/) {
   $perlver = $1;  # probably 5.something or 6.something
}


# Handle input parameter and open the file:

while (($#ARGV >= 0) && ($ARGV[0] =~ /^-[dviq]/)) {
   if ($ARGV[0] =~ /^-.*d/) {
      $use_dd = 1;
   }
   if ($ARGV[0] =~ /^-.*v/) {
      $verbose++;
   }
   if ($ARGV[0] =~ /^-.*i/) {
      $immediatew = 1;
      if ($ARGV[0] =~ /^-.*q/) {
	 $quick = 1;
      }
   };
   shift @ARGV;
}
print "dbg: use_dd=$use_dd, immediatew=$immediatew, quick=$quick, verbose=$verbose\n";

($drive2zero, $log_input, $script2zeroit) = @ARGV;

if ((0 > $#ARGV) || (!defined $log_input) || ((!defined $script2zeroit) && !$immediatew)) {
   die "ddrwipe.pl  (\"zero-out the GOOD sectors on a bad drive\")   v.$ver_ordinal_and_date\n" .
       "\n" .
       "Usage:\n" .
       "   perl ddrwipe.pl /dev/hdd drive.log script2zeroit\n" .
       "   ./script2zeroit\n" .
       "\n" .
       "or, to generate 'dd' commands instead of 'ddrescue' commands:\n" .
       "   perl ddrwipe.pl -d /dev/hdd drive.log script2zeroit\n" .
       "\n" .
       "or, to immediately write to the drive, instead of generating a script:\n" .
       "   perl ddrwipe.pl -i /dev/hdd drive.log\n" .
       "\n" .
       "or, to immediately write to the drive, but skip rest of block on errors:\n" .
       "   perl ddrwipe.pl -iq /dev/hdd drive.log\n" .
       "\n" .
       "or, for verbose output:\n" .
       "   perl ddrwipe.pl -v /dev/hdd drive.log script2zeroit\n" .
       "   ...\n";
     #  "\n" .
     #  "'-v' can be repeated for very verbose output.\n";
}

if ($verbose) {
   print "You are using Perl version $perlver\n";
   ($sec,$min,$hr,$day,$mon,$year) = localtime(time);
   printf "%4d-%02d-%02d %02d:%02d:%02d\n", $year+1900, $mon+1, $day, $hr, $min, $sec;
}


######################################################################
# Here's a sample of what a ddrescue log file looks like:
#
#    # Rescue Logfile. Created by GNU ddrescue version 1.1
#    #      pos        size  status
#    0x00000000  0x0004BE00  +
#    0x0004BE00  0x00000200  -
#    0x0004C000  0x33078E00  +
#    0x330C4E00  0x00001000  -
#    0x330C5E00  0x8ACDF200  +
#
######################################################################


# input is an integer represented as a string (probably hexadecimal preceeded by '0x'),
# result is the integer divided by 512.
sub divby_512 {
   local($dividend) = @_;
   # print "dividend=$dividend\n";
   local($result) = 0;
   if ($dividend =~ /^0x/) {
      # print "tricky way!\n";
      # tricky divide by 512: drop the last 2 hexidecimal digits, and divide by 2 (which is more robust for very large integers)
      $result = substr($dividend,0,length($dividend)-2);
      $result = int(eval($result) / 2);  # gotta use eval() because it starts with 0x
   } else {
      # print "conventional way,\n";
      $result = int((0+$dividend) / 512);
   }
   # print "result=$result\n";
   return $result;
}


# convert a really big positive integer into a string with commas
sub commas {
   local( $i ) = @_;
   local( $result, $x );
   if ($i <= 9999.5) {
      $result = sprintf( "%4.0f", $i );
   } else {
       $x = sprintf( "%1.0f", $i );
       $result = '';
       $x = reverse( $x );
       while ('' ne $x) {
          if ($result ne '') {
             $result .= ',';
          }
          $result .= substr( $x, 0, 3 );
          if (length($x) > 3) {
             $x = substr( $x, 3 );
          } else {
             $x = '';
          }
      }
      $result = reverse( $result );
   }
   # printf "dbg: %1.0f = %s\n", $i, $result;
   return $result;
}


$zeros = pack("c",0) x 512;  # a sector of zero-bytes


sub process1file {
   local( $drive2zero, $log_input, $script2zeroit ) = @_;
   local( $numranges, $total_num_sectors, %counts, @tmp, %tmp2, $pos_byt, $len_byt, $status, $len_sect, $bytes, $megabytes, $percentage, $unrecovered, $unrecovered2, $n, $dots, $succ, $fail );
   if (!open( LOG_INPUT, "<$log_input")) {
      print "ERROR: Could not read $log_input, $!\n";
      return;
   }
   if (!$immediatew) {
      if (!open( SCRIPT_OUTPUT, ">$script2zeroit")) {
	 print "ERROR: Could not write $script2zeroit, $!\n";
	 return;
      }
      binmode SCRIPT_OUTPUT;  # no CRs in output even if run under DOS/Windows
      print SCRIPT_OUTPUT "# This script will zero the good sectors on $drive2zero prior to warranty return.\n";
   }
   if ($immediatew) {
      if (!open( DRIVE2WIPE, "+>$drive2zero" )) {
	 print "ERROR: Could not open '$drive2zero' for writing, $!\n";
	 return;
      }
      binmode DRIVE2WIPE;
   }
   # read log_input
   $numranges = 0;
   $total_num_sectors = 0;
   $previous_status = ' ';
   $inplinenum = 0;
   $dots = 0;
   while (<LOG_INPUT>) {
      $inplinenum++;
      if (($_ =~ /\d/) && ($_ !~ /^\#/)) {
         $_ =~ s/\s*$//;  # strip trailing blanks, tabs, cr, lf, etc.
         # Parse a ddrescue log file line:
         @tmp = split( /\s+/, $_ );
         # for ($i=0; $i<=$#tmp; $i++) { print "dbg: tmp[$i]='" . $tmp[$i] . "'\n"; }
         ($pos_byt, $len_byt, $status) = @tmp;

         if (($#tmp != 2) || ($pos_byt !~ /\d/) || ($len_byt !~ /\d/)) {
            if ((1 == $#tmp) && (0 == $numranges) && ($len_byt =~ /^[\?\*\/\-\+]$/)) {
               # the new status line (introduced in ddrescue 1.6 or 1.7)
               $pos_sect = &divby_512($pos_byt);
               $status = $len_byt;
               print SCRIPT_OUTPUT "# IGNORED Status: ";
               if ($status eq '?') {
                  print SCRIPT_OUTPUT "copying data";
               } elsif ($status eq '*') {
                  print SCRIPT_OUTPUT "trimming bad areas";
               } elsif ($status eq '/') {
                  print SCRIPT_OUTPUT "splitting bad areas";
               } elsif ($status eq '-') {
                  print SCRIPT_OUTPUT "retrying bad blocks";
               } elsif ($status eq '+') {
                  print SCRIPT_OUTPUT "finished";
               } else {
                  print SCRIPT_OUTPUT "unrecognized status '$status'";
               }
               print SCRIPT_OUTPUT " (status code = '$status'), at sector $pos_sect\n";
            } else {
               print "ERROR: unrecognized input on line $inplinenum of '$log_input':\n$_\n";
               print SCRIPT_OUTPUT "# IGNORED Unrecognized input: $_\n";
            }
            next;
         }

         if ($len_byt eq '-1') {
            print "Note: length=-1, line $inplinenum of '$log_input' ignored:  $_\n";
	    print SCRIPT_OUTPUT "# IGNORED '-1' length: $_\n";
            next;
         }
         # convert byte offsets and counts to LBA sector number & sector count:
         $pos_sect = &divby_512($pos_byt);
         $len_sect = &divby_512($len_byt);

         if (!defined $counts{$status}) {
            $counts{$status} = 0;
            $numintervals{$status} = 0;
         }
         $counts{$status} += $len_sect;
         $total_num_sectors += $len_sect;
         $numranges++;
         if ($status ne $previous_status) {
            $numintervals{$status}++;
            $previous_status = $status;
         }
	 if ($immediatew) {
	    if (("+" eq $status) && ($len_sect > 0)) {
	       if ($dots) {
		  print "\n";
		  $dots = 0;
	       }
	       print "$_\n";
	       if (!seek( DRIVE2WIPE, $pos_sect*512.0, 0 )) {
	          print "ERROR: seek failed, ofs=$pos_byt, $!\n";
	       } else {
		  print " dbg: seek(," . ($pos_sect*512.0) . ",0)\n";
	          $n = 0.0;
		  $succ = $fail = 0;
		  $oldtime = (times)[0];
	          while ($len_sect--) {
		     $rc = syswrite( DRIVE2WIPE, $zeros, 512 );
		     # $rctmp = $rc;
		     # if (!defined $rc) {
		     #    $rctmp = "undef";
		     # }
		     # $rctmp .= '';
		     # if ('512' eq $rctmp) {
		     #    print "w";
		     # } else {
		     #    print " dbg: wr(),rc=$rctmp\n";
		     # }
		     if (!defined $rc) {
			$fail++;
			if ($quick) {
			   if ($dots) {
			      print "\n";
			      $dots = 0;
			   }	
		           print "ERROR: write error at ofs=$pos_byt+($n" . "x512), $!.\n";
		           last;
			}
			# attempt to recover from the error;
			close(DRIVE2WIPE);
			if (!open( DRIVE2WIPE, "+>$drive2zero" )) {
			   print "ERROR: Could not re-open '$drive2zero', $!\n";
			   return;
			}
			binmode DRIVE2WIPE;
			$n += 1.0;
			if (!seek( DRIVE2WIPE, ($pos_sect+$n)*512.0, 0 )) {
			   print "ERROR: seek failed, ofs=$pos_byt+($n" . "x512), $!.\n";
			   last;
			}
			# print " --err/recvry (n=$n-1)--\n"; #dbg
		     } else {
			$succ++;
			$n += 1.0;
		     }
		     $newtime = (times)[0];
		     if (   (($succ+$fail) == 1024)
		         || (($newtime-$oldtime) >= 2)
			 || ((0==$len_sect) && (($succ+$fail)>0))
		        ) {
		        if ($fail && !$succ) {
			   print "x";
		        } elsif ($succ && !$fail) {
			   print ".";
		        } elsif ($succ > $fail) {	
			   print "/";
			   # print "s=$succ,f=$fail "; #dbg
		        } else {
			   print "\\";
		        }
		        $dots++;
		        $succ = $fail = 0;
		        $oldtime = $newtime;
		        if (79 == $dots) {
		           print "\n";
		           $dots = 0;
		        }
		     }
	          }
	       }
            }
	 } else {
	    if ('+' eq $status) {
	       print SCRIPT_OUTPUT "# ZERO:  $_\n";
               if ($verbose) {
                  print SCRIPT_OUTPUT "# dbg:  $pos_byt / 512 = $pos_sect,  $len_byt / 512 = $len_sect\n";
               }
	       if ($use_dd) {
	          print SCRIPT_OUTPUT "dd bs=512 seek=$pos_sect count=$len_sect if=/dev/zero of=$drive2zero\n";
	       } else {
		  $pos_sect .= 'b';  # 'b' suffix means units are hardware blocks (512 bytes)
		  $len_sect .= 'b';
		  print SCRIPT_OUTPUT "ddrescue -c 1 -i 0 -o $pos_sect -s $len_sect /dev/zero $drive2zero ddrwipe.log\n";
	       }
	    } else {
	       print SCRIPT_OUTPUT "# SKIP:  $_\n";
	    }
         }
      }#if
   }#while
   close LOG_INPUT;

   if ($immediatew) {
      if ($dots) {
	 print "\n";
	 $dots = 0;
      }
      print "done.\n";
   } else {
      if (!$use_dd) {
	 print SCRIPT_OUTPUT "echo 'Done.  The logfile is ddrwipe.log; delete it to start over.'\n";
      }
      close SCRIPT_OUTPUT;
   }

   if ($verbose) {
      %tmp2 = %counts;
      foreach $status ('+','-','/',keys %counts) {
         if (defined $tmp2{$status}) {
            printf "$status%11d sectors =", $counts{$status};
            $bytes = &commas( $counts{$status} * 512.0 );
            $megabytes = $counts{$status} / 2048;
            $percentage = 100.0 * ($counts{$status} / $total_num_sectors);
            printf "%17s bytes =%10.2f MB = %6.3f%% of total", $bytes, $megabytes, $percentage;
            if ($verbose > 1) {
               printf ", in %4d intervals", $numintervals{$status};
            }
            print "\n";
            delete $tmp2{$status};
         }
      }
      printf "Total:\n" .
             "  %10d sectors =%17s bytes =%10.2f MB (in %d intervals)\n",
             $total_num_sectors, &commas($total_num_sectors*512.0), $total_num_sectors/2048, $numranges;
      $unrecovered2 = 0;
      if (defined $counts{'-'}) {
         $unrecovered2 += $counts{'-'};
      }
      if (defined $counts{'/'}) {
         $unrecovered2 += $counts{'/'};
      }
      if (defined $counts{'?'}) {
         $unrecovered2 += $counts{'?'};
      }
      $unrecovered = $total_num_sectors;
      if (defined $counts{'+'}) {
         $unrecovered -= $counts{'+'};
      }
      printf "Unrecovered:";
      if ($unrecovered2 != $unrecovered) {
         print "           (unrecovered sector count is $unrecovered, or maybe $unrecovered2)";
      }
      printf "\n" .
             "%12d sectors =%17s bytes =%10.2f MB = %6.3f%% of total\n",
             $unrecovered, &commas($unrecovered*512.0), $unrecovered/2048,
             100.0*($unrecovered/$total_num_sectors);
      print "\n"; #extra blank line at end
   }#if ($verbose)
   if (!$immediatew) {
      print "Created: '$script2zeroit' (run it to zero-out the data from $drive2zero)\n";
      `chmod 777 $script2zeroit`;
   }
}#sub process1file

&process1file( $drive2zero, $log_input, $script2zeroit );

__END__

