#!/usr/bin/perl -w

# FAT32INF.PL -- a program to display FAT32 file system information.
# (Written in Perl 4.036, and compatible with Perl 5.)
#
#
# Copyright 2007-2009, by David A. Burton.
# Burton Systems Software
# http://www.burtonsys.com/
# POB 4157, Cary, NC 27519-4157
# Tel: +1 919-481-0149
# Email: ncdave4life at gmail.com
#
#

$ver_ordinal_and_date = $b_bootstrap = $b_unused1 = $b_unused2 = $b_unused3 = 0;  # shut up Perl warnings

#  =>keyflag<=-- "%v, %d"
$ver_ordinal_and_date = "7, 12-May-15";
# the TLIB Version Control version number and last-modified date


# Here are a couple of descriptions of the FAT32 file system format:
#   http://www.maverick-os.dk/FileSystemFormats/FAT32_FileSystem.html
#   http://home.att.net/~rayknights/pc_boot/w95bboot.htm
#   http://home.no.net/tkos/info/fat32.html


$| = 1;   # predefined variable. If <> 0 then each print to the console
          # will immediatly be displayed, instead of buffered.

$copr = "Copr. 2007-2009, David A. Burton, Burton Systems Software.  www.burtonsys.com\n";
print "fat32inf.pl -- FAT32 File Sector Information Utility.   v.$ver_ordinal_and_date\n" . $copr;

$timer{'main'} = 0.0;
$timer{'main'} -= (times)[0];

$o_verbose = 0;  # '-v' sets it to 1, -vv or -v2 or "-v -v" sets it to 2
$o_debugmode = 0;  # '-D' sets it to 1, -DD or -D2 or "-D -D" sets it to 2
$o_minmem = 0;  # '-m' sets it to 1
$o_showbootsect = 0;  # '-b' sets it to 1
$o_showFAT = 0;  # '-f' sets it to 100, '-F' sets it to 1000000
$o_listfiles = 0;  # '-l' sets it to 1, '-ll' or '-l2' sets it to 2
$o_rawlistfiles = 0;  # '-L' sets it to 1, '-LL' or '-L2' sets it to 2, '-L3' sets it to 3
$o_longnames = 0;  # '-n' sets it to 1
$o_search4bootsect = 0;  # '-h' sets it to 1
$o_lookup_sector = 0;  # '-s' sets it to 1
$o_listdeletedfiles = 0;  # '-d' sets it to 1
$o_exportFAT2log = 0;  # '-e' sets it to 1
$forced_FAT_start_sect = 0;  # '-f=95' sets it to 95
$o_reconcile = 0;  # '-r' sets it to 1
$o_reconcilewrite = 0;  # '-R' sets it to 1
$o_correctFATfromDirs = 0;  # '-c' sets it to 1
$o_pause = 0;  # '-p' sets it to 1

# initialize globals:
$lba_of_bootsect = -1;  # which boot sector is currently loaded
$fat = $fat2 = '';
$fat1fixes = $fat2fixes = 0;

$four0bytes = "\0\0\0\0";
$thirtytwo0s = ("\0" x 32);  # null directory entry (pads out the last directory sector)
$thirtyone0s = "\xE5" . ("\0" x 31);  # alternate null directory entry
$zerosect = "\0" x 512;
$badclustermark = pack( "L", 0x0FFFFFF7 );
if ("\xF7\xFF\xFF\x0F" ne $badclustermark) {
   print "ERROR: This program requires a 32-bit Perl on a little-endian machine.\n";
   &quit(3);
}


use Config;
$uselargefiles = '';
if (defined $Config{'uselargefiles'}) {
   $uselargefiles = $Config{'uselargefiles'};
}
# printf "dbg:  uselargefiles = '$uselargefiles'\n";
if ('' eq $uselargefiles) {
   print "Warning: this old version of Perl does not support files > 2GB (or maybe 4GB?) in size.\n";
   if ($o_exportFAT2log > 2) {
      print "ERROR: To use the '-e3' option with this program you MUST use a version of Perl\n" .
            "which was compiled with 'uselargefiles' defined.\n";
      &quit(3);
   }
}


# Perl pack/unpack format representing the structure of the FAT32 Boot Sector:
$BootSector_fmt =
  'a3'  . # 000h  3 bytes  JMP 0x80h
  'a8'  . # 003h  8 bytes  Oem ID - Name of the formatting OS
          ## BIOSParameter Block (00Bh - 03Fh):
  'S'   . # 00Bh  2 bytes  Bytes per Sector on the physical medium - 512/1024/2048/4096, but normally 512 bytes
  'C'   . # 00Dh  1 bytes  Sectors per Cluster - 1, 2, 4, 8, 16, 32, 64 or 128 sectors
  'S'   . # 00Eh  2 bytes  Reserved sectors in front of the FAT(s) incl. the Boot sector (usually 32 for FAT32, 1 for FAT12FAT16, 0 for NTFS)
  'C'   . # 010h  1 bytes  Number of FAT copies - Normally 2 for FAT, 0 for NTFS
  'a4'  . # 011h  4 bytes  Not used in FAT32 or NTFS (1st 2 bytes are number of root directory entries in FAT12/FAT16, 2nd 2 bytes can be a 16-bit total number of data sectors for FAT12/FAT16, if that value fits in 16 bits)
  'C'   . # 015h  1 bytes  Media Descriptor - The same as in FAT16, but FAT32 is only allowed on harddrives, so the value is F8h
  'S'   . # 016h  2 bytes  Not used in FAT32 or NTFS (number of sectors/fat in FAT12/FAT16)
  'S'   . # 018h  2 bytes  Sectors per Track - The disc geometry used when formatting the partition
  'S'   . # 01Ah  2 bytes  Heads - The disc geometry used when formatting the partition
  'L'   . # 01Ch  4 bytes  Supposedly the number of "hidden sectors," but actually usually the sector number of the partition boot sector relative to the partition table sector that points to the partition.  Usually 63 for extended partitions, and for the first primary partition; usually larger for other primary partitions.
  'L'   . # 020h  4 bytes  Number of sectors in the partition (for FAT32 and some FAT16 partitions)
  # Note: offsets 24-64 are defined for FAT32 only:
  'L'   . # 024h  4 bytes  Number of sectors per FAT (FAT32 only)
  'S'   . # 028h  2 bytes  FAT32 handling flags
  'S'   . # 02Ah  2 bytes  FAT32 Drive Version (High byte = Major version, Low byte = Minor version)
  'L'   . # 02Ch  4 bytes  Cluster number for the start of the FAT32 Root Directory Table (usually 2)
  'S'   . # 030h  2 bytes  Sector number of File System Information Sector, BIGFATBOOTFSINFO struct
  'S'   . # 032h  2 bytes  Sector number of Backup Boot Sector
  'a12' . # 034h 12 bytes  Reserved
          ## Extended BIOS Parameter Block (040h - 059h):
  'C'   . # 040h  1 bytes  Logical Drive Number - Normally 00h for floppies and 80h for hard drives.
  'C'   . # 041h  1 bytes  Current Head
  'C'   . # 042h  1 bytes  Signature byte, should be 29h to indicate that Serial Number, Volume Label, and System ID are valid
  'L'   . # 043h  4 bytes  Serial Number (randomly generated)
  'a11' . # 047h 11 bytes  Volume Label - The same as stored in a special file in the root directory.
  'a8'  . # 052h  8 bytes  System ID - This is the string 'FAT32 '
          ## Code:
  'a420'. # 05Ah 420 bytes possibly bootstrap loader codeb_unused2
  'S';    # 1FEh  2 bytes  Executable sector signature (AA55h)


# Unpack a boot sector using $BootSector_fmt as the template:
sub unpack_a_boot_sector {
  local( $onesector ) = @_;
  local( @results ) = unpack( $BootSector_fmt, $onesector );
  return @results;
}


# display a call-traceback diagnostic, for debugging
sub traceback {
    print "***** Traceback (most recent call first):\n";
    local($lvl) = 0;
    local($package) = local($filename) = '';
    local($line) = 0;
    while (1) {
       ($package, $filename, $line) = caller($lvl);
       if ((!defined $filename) || ('' eq $filename)) {
          last;
       }
       if ('main' ne $package) {
          print " pkg='$package' ";
       }
       print "called from $filename, line $line\n";
       $lvl++;
    }
    print "***** traceback done.\n";

   #  my $frame = 2;
   #  while (my ($package, $filename, $line, $sub) = caller($frame++)) {
   #      if ($package eq 'main') {
   #          $sub =~ s/main:://;
   #          push(@trace, "  $sub() called from $filename on line $line")
   #      } else {
   #          push(@trace, "  $package::$sub() called from $filename on line $line\n");
   #      }
   #  }
   #  print STDERR join("\n", reverse @trace)."\n";
   #  print STDERR "Error: $err"; #already has \n at end
   #  exit(1);
} # traceback


# determine and display RAM usage for this program (only works under Linux)
sub get_stat_info {
  local(@stat);
  # open and read the stat file for this process
  if (!open(STAT, "</proc/$$/stat")) {
     print "ERROR: Couldn't open /proc/$$/stat, $!\n";
  } else {
     local(@stat) = split( /\s+/ , <STAT> );
     close STAT;
     $vsize = int( $stat[22] / 1024 );  # virtual memory size in kb
     $rss = $stat[23] * 4;  # working set size in kb
     printf "virtual memory used = %s KB,  working set (rss) = %s KB  (from /proc/$$/stat)\n", &commas5($vsize), &commas5($rss);
  }
}


# Exit the program.
# First parameter is errorlevel set via the 'exit' function; if unspecified '2' is used.
# If second parameter is specified as '1' then a traceback is done, too.
# If '-p' was specified ($o_pause is set) then the program pauses and awaits a keypress before exiting.
sub quit {
   local($errorlevel, $tracebk) = @_;
   local($junk);
   if ((defined $tracebk) && $tracebk) {
      &traceback;
   }
   if (!defined $errorlevel) {
      $errorlevel = 2;
   }
   if ($o_debugmode) {
      print "Done.\n";
   }
   &get_stat_info;
   $timer{'main'} += (times)[0];
   printf "runtime = %.2f seconds.\n", $timer{'main'};
   if ($o_pause) {
      print "Exiting program, errorlevel=$errorlevel.  Press [Enter]: ";
      $junk = getc;
      print "\n";
   }
   exit $errorlevel;
} # sub quit


# @sv_ARGV = @ARGV;  # save raw parameter list, for display to the user

sub do1option {
  local($ch,$valu) = @_;
  if ($o_debugmode || ('D' eq $ch)) {
     print "dbg: do1option('$ch',$valu);\n";
  }
  if ('v' eq $ch) {
     $o_verbose += $valu;
  } elsif ('D' eq $ch) {
     $o_debugmode += $valu;
  } elsif ('m' eq $ch) {
     $o_minmem += $valu;
  } elsif ('b' eq $ch) {
     $o_showbootsect += $valu;
  } elsif ('f' eq $ch) {
     $o_showFAT += 100;
     if ($valu > 1) {
        $forced_FAT_start_sect = $valu;
     }
  } elsif ('F' eq $ch) {
     $o_showFAT += 1000000;
  } elsif ('l' eq $ch) {
     $o_listfiles += $valu;
  } elsif ('L' eq $ch) {
     $o_rawlistfiles += $valu;
  } elsif ('n' eq $ch) {
     $o_longnames += $valu;
  } elsif ('d' eq $ch) {
     $o_listdeletedfiles += $valu;
  } elsif ('h' eq $ch) {
     $o_search4bootsect += $valu;
  } elsif ('s' eq $ch) {
     $o_lookup_sector += $valu;
  } elsif ('e' eq $ch) {
     $o_exportFAT2log += $valu;
  } elsif ('r' eq $ch) {
     $o_reconcile += $valu;
  } elsif ('R' eq $ch) {
     $o_reconcilewrite += $valu;
  } elsif ('c' eq $ch) {
     $o_correctFATfromDirs += $valu;
  } elsif ('p' eq $ch) {
     $o_pause += $valu;
  } elsif ('-' ne $ch) {
     print "Warning: unrecognized option (\"-$ch\") ignored.\n";
  }
  # Note: if you add another option, be sure to fix the 'while' condition, too
} # sub do1option


# Check for command-line options.
# These are equivalent (where x, y, and z are any legal options):
#    -x -y -z
#    -xyz
#    -x1 -y1 -z1
# These are also equivalent:
#    -xx
#    -x2
# Etc.
while (($#ARGV >= 0) && ($ARGV[0] =~ /^-[vdDmbfFlLsherRncp]/)) {
   if ($ARGV[0] =~ /^-.(\=|)\d/) {
      $ch = substr( $ARGV[0], 1, 1 );
      $valu = substr( $ARGV[0], 2 );
      if ($valu =~ /^\=/) {
         $valu = substr( $valu, 1 );
      }
      $valu = 0 + $valu;
      # print "dbg: ch='$ch', valu='$valu'\n";
      &do1option( $ch, $valu );
   } else {
      foreach $ch (split(//,$ARGV[0])) {
         &do1option( $ch, 1 );
      }#foreach
   }
   shift @ARGV;
}


if ($o_debugmode) {
   print "dbg: o_showFAT = $o_showFAT\n";
}


$numargs = $#ARGV + 1;
if ($o_debugmode) {
   print "dbg: after recognized options processed, numargs=$numargs, args='" . join(" ",@ARGV) . "'\n";
}
if ($o_search4bootsect) {
   if ($numargs != 3) {
      print "ERROR: -h option requires exactly 3 additional parameters, not $numargs.\n";
      &quit;
   }
} elsif ($o_lookup_sector) {
   if ($numargs < 3) {
      &givehelp;
   }
} elsif ($numargs != 2) {
   &givehelp;
}
if ($o_longnames && !$o_listfiles) {
   $o_listfiles = 1;
}
if ($o_search4bootsect && ($o_lookup_sector || $o_listfiles || $o_rawlistfiles)) {
   print "ERROR: You may not specify -h with -l, -L, -n, or -s\n";
   &quit;
}
if ($o_minmem && ($o_reconcile || $o_reconcilewrite || $o_correctFATfromDirs)) {
   print "ERROR: You may not specify -m with -R, -r, or -c\n";
   &quit;
}

sub givehelp {
   print 'ERROR: $#ARGV = ' . $#ARGV . "\n\n";
   print "Usage (basic info mode):\n" .
         "  perl fat32inf.pl {options} drive bootsect\n" .
         "or (search for boot sectors):\n" .
         "  perl fat32inf.pl -h drive firstsect lastsect\n" .
         "or (look up files from sector numbers):\n" .
         "  perl fat32inf.pl -s drive bootsect sect1 sect2 sect3...\n" .
         "\n" .
         "Where:\n" .
         "  'drive' is a device name or a file containing a drive image, and\n" .
         "  'bootsect' is the LBA sector number (0-based) of the partition boot sector\n" .
         "  'firstsect' and 'lastsect' specify a range of sector numbers; those which\n" .
         "           resemble boot sectors will be displayed\n" .
         "  'sect1', 'sect2', etc. are LBA sector numbers to report the containing files for.\n" .
         "  {options} are optional:\n" .
         "    -m    reduced Memory usage (usually makes little difference)\n" .
         "    -s    lookup filenames from LBA Sector numbers (prefix with '0x' for hexidecimal)\n" .
         "    -h    Heuristic search for boot sectors\n" .
         "    -b    display Boot sector in human-understandable format\n" .
         "    -f    display the first 100 sectors of the FAT(s) (-ff for 200, etc.)\n" .
         "    -f=95 same as -f, but dump FAT starting at LBA 95 instead of location in boot sector\n" .
         "    -F    display the entire main FAT (probably very long!)\n" .
         "    -e    export FAT #1 to ddrescue-format logfiles (sector lists)\n" .
         "    -e2 (or -ee)  export 0-sectors from FAT contents as well as what FAT points to\n" .
         "    -e3   reconstruct lost ddrescue logfile by examining FAT \& checking used clusters for 0-sectors (SLOW!)\n" .
         "    -r    reconcile differences between the two FATs, but don't write 'em to disk\n" .
         "    -R    reconcile differences between the two FATs and write 'em to disk\n" .
         "    -c    correct FAT errors from directory information\n" .
         "    -l    List all files, regular format (note: '-l' is lower-case '-L', not '-1')\n" .
         "    -l2 (or -ll)  List all files, plus full cluster list for each file\n" .
         "    -l3   List files + cluster list + cluster list terminator for each file\n" .
         "    -n    add long file Names to the '-l' list output\n" .
         "    -L    List all files, raw directory format\n" .
         "    -L2 (or -LL)  List all files + long name fragments, raw directory format\n" .
         "    -d    List deleted files, raw directory format\n" .
         "    -d2 (or -dd)  List deleted files + long name fragments, raw directory format\n" .
         "    -v    Verbose mode\n" .
         "    -D    Debug mode prints (or '-D2' or '-DD' for even more debug prints)\n" .
         "    -p    Pause before exiting\n" .
         "\n" .
         "(Note: often the main boot sector is sector 63, and the backup boot\n" .
         "sector is sector 69.)\n" .
         "\n" .
         "Examples:\n" .
         "  perl fat32inf.pl -b /dev/sdb1 0           (display partition boot sector)\n" .
         "  perl fat32inf.pl -b /dev/sdb 63           (same, if sdb1 is at LBA 63)\n" .
         "  perl fat32inf.pl -bl /dev/hdb1 0          (display boot sector and list files)\n" .
         "  perl fat32inf.pl -b -l /dev/hdb1 0        (same as previous example)\n" .
         "  perl fat32inf.pl -s /dev/hdb1 0 0x19E6    (look up sector #6330, show file name)\n" .
         "  perl fat32inf.pl -s /dev/hdb 63 6630      (same, if hdb1 is at LBA 63)\n" .
         "  perl fat32inf.pl -s drive.ima 63 0x18E6   (same, if drive.ima is copy of /dev/hdb)\n" .
         "  perl fat32inf.pl -s drive.ima 63 6630 19276 22040  (look up 3 sectors at once)\n" .
         "\n"
        # . $copr
         ;
   &quit(1);
} # sub givehelp


# input is a string containing a decimal or hexidecimal positive integer
# output is the integer
sub hex_or_positive_decimal {
  local( $str ) = @_;
  local( $result ) = 0;
  if ($str =~ /^[0-9]+$/) {
     $result = 0+$str;
  } elsif ($str =~ /^0x[0-9a-fA-F]+$/) {
     $result = eval $str;
  } else {
     print "ERROR: '$str' is not a positive integer or C-format hexidecimal integer.\n";
     &quit(2,1);
  }
  if ($o_debugmode) {
     print "dbg: '$str' = $result\n";
  }
  return $result;
} # sub hex_or_positive_decimal


# input is an integer or string of digits
# output is the string of digits with commas added for readability (if it is 4 digits or more)
sub commas {
   local($x) = shift;
   local($tmp) = $x;
   $x .= "";  # make sure it's a string
   local($result) = "";
   while (length($x) > 3) {
      # print "dbg:  x='$x'  substr(x,-3)='" . substr($x,-3) . "'  substr(x,0,len-3)='" . substr( $x, 0, length($x)-3 ) ."'\n";
      if (length($result) > 0) {
         $result = "," . $result;
      }
      $result = substr($x,-3) . $result;
      $x = substr( $x, 0, length($x)-3 );
   }
   if (length($x) > 0) {
      if (length($result) > 0) {
         $result = "," . $result;
      }
      $result = $x . $result;
   }
   # print "dbg:  '$tmp' = '$result'\n";
   return $result;
} # sub commas


# equivalent to &commas except that it only adds commas if the number is 5 digits or more
sub commas5 {
   local($x) = shift;
   local($result);
   $x .= "";
   if (length($x) > 4) {
      $result = &commas($x);
   } else {
      $result = $x;
   }
   return $result;
}


# some command-line parameter processing:

# first two parameters are common to all modes:
$drive = $ARGV[0];
$lba_offset1 = $ARGV[1];
$lba_offset1 = &hex_or_positive_decimal( $lba_offset1 );
# handle 3rd parameter for '-h' mode
if ($o_search4bootsect) {
   # '-h' (heuristic search for boot sectors)
   $lba_offset2 = $ARGV[2];
   $lba_offset2 = &hex_or_positive_decimal( $lba_offset2 );
   if ($lba_offset2 < $lba_offset1) {
      print "ERROR: 2rd parameter (\"$lba_offset1\") must be <= 3nd parameter (\"$lba_offset2).\n";
      &quit;
   }
} else {
   undef $lba_offset2;
}


# unpack the interesting fields of a boot sector into global variables
sub unpack_boot_sect_to_globals {
  local( $fbuf ) = @_;
  local(@bootsectorfields) = &unpack_a_boot_sector( $fbuf );
  (  $b_jmp80                 , # 000h  3 bytes  JMP 0x80h
     $b_OemId                 , # 003h  8 bytes  Oem ID - Name of the formatting OS
                                ## BPB: BIOS Parameter Block (00Bh - 03Fh):
     $b_bytesPerSector        , # 00Bh  2 bytes  Bytes per Sector on the physical medium - Normally 512 bytes
     $b_sectorsPerCluster     , # 00Dh  1 bytes  Sectors per Cluster - 1, 2, 4, 8, 16, 32, 64 or 128 sectors
     $b_resSectBeforeFAT      , # 00Eh  2 bytes  Reserved sectors in front of the FAT(s) incl. the Boot sector (usually 32 for FAT32, 0 for NTFS)
     $b_numFATs               , # 010h  1 bytes  Number of FAT copies - Normally 2
     $b_unused1               , # 011h  4 bytes  Not used in FAT32 or NTFS
     $b_F8h                   , # 015h  1 bytes  Media Descriptor - The same as in FAT16, but FAT32 is only allowed on harddrives, so the value is F8h
     $b_unused2               , # 016h  2 bytes  Not used in FAT32 or NTFS
     $b_sectorsPerTrack       , # 018h  2 bytes  Sectors per Track - The disc geometry used when formatting the partition
     $b_heads                 , # 01Ah  2 bytes  Heads - The disc geometry used when formatting the partition
     $b_numHiddenSectors      , # 01Ch  4 bytes  supposedly the number of sectors on the disk from start of the partition to beginning of the first FAT, but actually the sector no. of the partition boot sector relative to the partition table sector that points to it
     $b_sectorsInPartition    , # 020h  4 bytes  Number of sectors in the partition
                                ## FAT32-only BPB fields:
     $b_sectorsPerFAT         , # 024h  4 bytes  Sectors per FAT
     $b_handlingFlags         , # 028h  2 bytes  FAT handling flags
     $b_driverVersion         , # 02Ah  2 bytes  FAT32 Driver Version (High byte = Major version, Low byte = Minor version)
     $b_clusterOfRootDirStart , # 02Ch  4 bytes  Cluster number for the start of the Root Directory
     $b_sectorOfFileSystInf   , # 030h  2 bytes  Sector number of File System Information Sector, BIGFATBOOTFSINFO struct
     $b_sectorOfBackupBootSect, # 032h  2 bytes  Sector number of Backup Boot Sector (often 6)
     $b_unused3               , # 034h 12 bytes  Reserved
                                ## Extended BIOS Parameter Block (040h - 059h):
     $b_logicalDriveNum       , # 040h  1 bytes  Logical Drive Number - Normally 00h for floppies and 80h for hard drives.
     $b_currentHead           , # 041h  1 bytes  Current Head
     $b_signature1            , # 042h  1 bytes  Signature (should be 29h)
     $b_serialNumber          , # 043h  4 bytes  Randomly generated serial number
     $b_volumeLabel           , # 047h 11 bytes  Volume Label - The same as stored in a special file in the root directory.
     $b_systemID              , # 052h  8 bytes  System ID - This is the string 'FAT32 '
                                ## Code:
     $b_bootstrap             , # 05Ah 420 bytes possibly bootstrap loader code
     $b_AA55h                   # 1FEh  2 bytes  Executable sector signature (AA55h)
  ) = @bootsectorfields;

  # Hack for Gregory Johnson's drive
  # $b_sectorsPerCluster = 64;  # temp dbg

} # sub unpack_boot_sect_to_globals


# Calculate various relative sector numbers, which will come in handy later.
# We assume that &unpack_boot_sect_to_globals() has already been called, and
# that $lba_of_bootsect has already been set.
sub calc_offsets {
  # Calculate the sector number of the first sector of the first data cluster
  # (which is cluster number 2, oddly enough), relative to the boot sector:
  $rel_first_data_sector = $b_resSectBeforeFAT + ($b_numFATs * $b_sectorsPerFAT);
  # (Note: for FAT12 & FAT16 you'd have to also add the root directory size,
  # because the FAT12/16 root directory is not stored in the data clustors.)

  # Calculate the data-area-base sector number, relative to the boot sector.
  # It would be the starting sector number of cluster 0, if there were such a
  # thing as cluster 0.
  $rel_sect_of_cluster0 = $rel_first_data_sector - (2 * $b_sectorsPerCluster);
  if ($o_debugmode) {
     print "dbg: b_resSectBeforeFAT=$b_resSectBeforeFAT,  b_numFATs=$b_numFATs, b_sectorsPerFAT=$b_sectorsPerFAT.\n";
     print "     rel_first_data_sector = b_resSectBeforeFAT + (b_numFATs * b_sectorsPerFAT) = $b_resSectBeforeFAT+($b_numFATs*$b_sectorsPerFAT) = $rel_first_data_sector.\n";
     print "dbg: rel_sect_of_cluster0 = rel_first_data_sector-(2*b_sectorsPerCluster) = $rel_first_data_sector-(2*$b_sectorsPerCluster) = $rel_sect_of_cluster0.\n";
  }

  # Calculate the locations of the two FATs:
  $rel_sect_of_FAT1 = $b_resSectBeforeFAT;

  # if command-line "-f=..." option was used to override FAT location, then use that:
  if ($forced_FAT_start_sect > 0) {
     if (&lba_to_rel($forced_FAT_start_sect) == $rel_sect_of_FAT1) {
        print "Note: specified FAT location is FAT #1 location from boot sector\n";
     } elsif (&lba_to_rel($forced_FAT_start_sect) == $rel_sect_of_FAT2) {
        print "Note: specified FAT location is FAT #2 location from boot sector\n";
     } else {
        print "Note: specified FAT location differs from FAT locations in boot sector\n";
     }
     $rel_sect_of_FAT1 = &lba_to_rel($forced_FAT_start_sect);
     print "specified FAT location = LBA $forced_FAT_start_sect = rel " . &lba_to_rel($forced_FAT_start_sect) . "\n";
     $b_numFATs = 1;  # ignore other FAT
  }

  if ($b_numFATs >= 2) {
     $rel_sect_of_FAT2 = $rel_sect_of_FAT1 + $b_sectorsPerFAT;
  } else {
     $rel_sect_of_FAT2 = $rel_sect_of_FAT1;
  }

  # Calculate number of data clusters:
  if ($b_sectorsPerCluster <= 0) {
     $num_data_clusters = $highest_cluster_num = 0;
  } else {
     $num_data_clusters = int( ($b_sectorsInPartition - $rel_first_data_sector) / $b_sectorsPerCluster );
     $highest_cluster_num = $num_data_clusters + 1;  # $num_data_clusters numbered 2 .. $highest_cluster_num
     if ($o_debugmode || !$o_search4bootsect) {
        printf "dbg: b_sectorsInPartition=%s  rel_first_data_sector=%s  bsectorsPerCluster=$b_sectorsPerCluster  num_data_clusters=%s\n",
           &commas($b_sectorsInPartition), &commas($rel_first_data_sector), &commas($num_data_clusters);
     }
  }

  local( $lastdatasect ) = (&cluster_to_lba($highest_cluster_num)) + ($b_sectorsPerCluster - 1);
  if ($highest_cluster_num >= (128 * $b_sectorsPerFAT)) {
     if ($o_debugmode || !$o_search4bootsect) {  # ($o_debugmode)
        print "Warning: highest cluster number is too large for the FAT size.\n";
        print "Adjusting it down from $highest_cluster_num to";
     }
     $highest_cluster_num = int($b_sectorsPerFAT * 128);
     if (!$o_search4bootsect) {  # ($o_debugmode) {
        print " $highest_cluster_num\n";
     }
  }
  if ($o_debugmode || !$o_search4bootsect) {  # $o_debugmode) {
     printf "highest_cluster_num=%s   lastdatasect= LBA %s   b_sectorsInPartition=%s\n",
            &commas5($highest_cluster_num), &commas5($lastdatasect), &commas5($b_sectorsInPartition);
  }
} # sub calc_offsets


# Test a sector to see how much it resembles a partition boot sector.
# Input is the LBA sector number of possible boot sector.
# Output is a score, 0..10, with 10 being definitely a boot sector and
# zero being not at all resembling a boot sector.
sub test4bootsect {
  local( $lba_offset ) = @_;
  local( $score ) = 0;  # result: an indication of the liklihood that it is a boot sector
  &ensure_bootsect_loaded( $lba_offset );
  if ($b_AA55h == 0x0AA55) {
     $score += 2.7;
  }
  if ($b_jmp80 =~ /^[\xEB\xE9]/) {
     # jmp instruction can begin with 0EBh or 0E9h
     $score += 0.5;
  }
  if (($b_numFATs >= 0) || ($b_numFATs <= 2)) {
     $score += 0.3;
  }
  if ($b_F8h == 0xF8) {
     $score += 0.8;
  }
  if ($b_bytesPerSector == 512) {
     # usual case
     $score += 1.4;
  } elsif (($b_bytesPerSector == 1024) || ($b_bytesPerSector == 2048) || ($b_bytesPerSector == 4096)) {
     # legal but rare
     $score += 1.2;
  }
  if (($b_sectorsPerCluster == 1) || ($b_sectorsPerCluster == 2) || ($b_sectorsPerCluster == 4) ||
      ($b_sectorsPerCluster == 8) || ($b_sectorsPerCluster == 16) || ($b_sectorsPerCluster == 32) ||
      ($b_sectorsPerCluster == 64) || ($b_sectorsPerCluster == 128)) {
     $score++;
  }
  if (($b_sectorOfBackupBootSect == 6) || ($b_sectorOfBackupBootSect == 0)){
     # seems to usually be 6 for FAT32 but 0 for NTFS
     $score += 0.3;
  }
  if ($b_OemId =~ /^NTFS/) {
     $score += 2.5;
     if ($b_resSectBeforeFAT == 0) {
        $score += 0.5;
     }
     if ($b_sectorsPerCluster == 8) {
        $score += 0.5;
     } else {
        $score -= 0.5;
     }
  } else {
     if ($b_OemId =~ /^(MSWIN|IBM|MSDOS)/) {
        $score += 0.5;
     }
     if ($b_systemID =~ /^FAT32/) {
        $score++;
     }
     if ($b_signature1 == 0x29) {
        $score++;
     }
     if ($b_resSectBeforeFAT == 1) {
        # usual value for FAT12/16 is 1
        $score += 0.1;
     } elsif ($b_resSectBeforeFAT == 32) {
        # usual value for FAT32 is 32
        $score += 0.5;
     }
     if ($highest_cluster_num < 65526) {
        $score -= 1.0;  # too small to be a legal FAT32 volume.
        if (!$o_search4bootsect) {  # $o_debugmode) {
           print "Warning: This is too small to be a legal FAT32 volume.  Highest cluster number is only $highest_cluster_num, but it should be >= 65526.\n";
        }
     }
  }
  return $score;
} # sub test4bootsect


# Display FAT32 file system info gleaned from a partition boot sector.
# Inputs are:
#   $fbuf             - a 512-byte sector, which might be a boot sector
#   $lba_offset       - the LBA disk address from which $fbuf was read
sub display_bootsect {
  local( $lba_offset ) = @_;
  local( $fileofs ) = (512 * $lba_offset);
  local( $hexfileofs ) = sprintf( "%X", $fileofs );
  if ('FFFFFFFF' eq $hexfileofs) {
      $hexfileofs = "";  # too big, can't display it
  } else {
      $hexfileofs = "= 0x0" . $hexfileofs;
  }
  printf "LBA $lba_offset = 0x0%X  -->  File offset of boot sector is: $fileofs $hexfileofs\n", $lba_offset;
  &ensure_bootsect_loaded( $lba_offset );
  printf "000h  3 bytes  Jmp                    = %02X %02X %02X\n", ord(substr($b_jmp80,0,1)), ord(substr($b_jmp80,1,1)), ord(substr($b_jmp80,2,1));
  printf "003h  8 bytes  OemId                  = '%-8s' (Oem ID - Name of formatting OS)\n", $b_OemId;
  printf "--- BIOSParameter Block (00Bh - 03Fh):\n";
  printf "00Bh  2 bytes  bytesPerSector         = %-6d     (Bytes per Sector on physical medium - Normally 512 bytes)\n", $b_bytesPerSector;
  printf "00Dh  1 bytes  sectorsPerCluster      = %-4d       (Sectors per Cluster - 1, 2, 4, 8, 16, 32, 64 or 128 sectors)\n", $b_sectorsPerCluster;
  printf "00Eh  2 bytes  resSectBeforeFAT       = %-6d     (Reserved sectors in front of FAT(s) incl. Boot sector)\n", $b_resSectBeforeFAT;
  printf "010h  1 bytes  numFATs                = %-4d       (Number of FAT copies - Normally 2 for FAT, 0 for NTFS)\n", $b_numFATs;
  printf "015h  1 bytes  F8h                    = 0x%02X       (Media Descriptor - should be F8h)\n", $b_F8h;
  printf "018h  2 bytes  sectorsPerTrack        = %-6d     (Sectors per Track - disc geometry used when formatting partition)\n", $b_sectorsPerTrack;
  printf "01Ah  2 bytes  heads                  = %-6d     (Heads - disc geometry used when formatting partition)\n", $b_heads;
  printf "01Ch  4 bytes  numHiddenSectors       = %-10d (sector no. of partition boot sector relative to the partition table sector that points to it)\n", $b_numHiddenSectors;
 if ('NTFS    ' eq $b_OemId) {
    printf "020h  4 bytes  unused for NTFS        = 0x%08X\n", $b_sectorsInPartition;
    printf "024h  4 bytes  unused for NTFS        = 0x%08X\n", $b_sectorsPerFAT;
    printf "028h  8 bytes  total sectors          = 0x%08X%04X%04X", $b_clusterOfRootDirStart, $b_driverVersion, $b_handlingFlags;
    $b_sectorsInPartition = ($b_clusterOfRootDirStart << 32) + ($b_driverVersion << 16) + $b_handlingFlags;
      printf " = %d\n", $b_sectorsInPartition;
    printf "0x30  8 bytes  clust num of \$MFT\n";
    printf "0x38  8 bytes  clust num of \$MFTMirr\n";
    printf "0x40  4 bytes  Clusters Per File Record Segment\n";
    printf "0x44  4 bytes  Clusters Per Index Block\n";
    printf "0x48  8 bytes  Volume Serial Number\n";
    printf "0x50  4 bytes  Checksum\n";
 } else {
  printf "020h  4 bytes  sectorsInPartition     = %-10d (Number of sectors in partition)\n", $b_sectorsInPartition;
  printf "024h  4 bytes  sectorsPerFAT          = %-8d   (Sectors per FAT)\n", $b_sectorsPerFAT;
  printf "028h  2 bytes  handlingFlags          = 0x%04X     (FAT handling flags)\n", $b_handlingFlags;
  printf "02Ah  2 bytes  driverVersion          = 0x%04X     (FAT32 Driver Version (High byte = Major version, Low byte = Minor version))\n", $b_driverVersion;
  printf "02Ch  4 bytes  clusterOfRootDirStart  = %-6d     (Cluster number for start of Root Directory, usually 2)\n", $b_clusterOfRootDirStart;
  printf "030h  2 bytes  sectorOfFileSystInf    = %-4d       (Sector number of File System Information Sector, BIGFATBOOTFSINFO struct)\n", $b_sectorOfFileSystInf;
  printf "032h  2 bytes  sectorOfBackupBootSect = %-4d       (Sector number of Backup Boot Sector)\n", $b_sectorOfBackupBootSect;
  printf "--- Extended BIOS Parameter Block (040h - 059h):\n";
  printf "040h  1 bytes  logicalDriveNum        = 0x%02X       (Logical Drive Number - Normally 00h for floppies and 80h for hard drives.)\n", $b_logicalDriveNum;
  printf "041h  1 bytes  currentHead            = %-4d       (Current Head)\n", $b_currentHead;
  printf "042h  1 bytes  signature1             = 0x%02X       (Signature - should be 29h)\n", $b_signature1;
  printf "043h  4 bytes  serialNumber           = 0x%08X (ID - Random generated serial number)\n", $b_serialNumber;
  printf "047h 11 bytes  volumeLabel            = '%-11s' (Volume Label - same as stored in a special file in root directory.)\n", $b_volumeLabel;
  printf "052h  8 bytes  systemID               = '%-8s' (System ID - Should be the string 'FAT32 ')\n", $b_systemID;
 }
  printf "--- Signature:\n";

  printf "1FEh  2 bytes  AA55h                  = 0x%04X     (Executable sector signature, should be 0xAA55h)\n", $b_AA55h;
 if ('NTFS    ' ne $b_OemId) {
  printf "Calculated highest cluster = %s = 0x0%X, last data sector = LBA %s\n",
         &commas($highest_cluster_num), $highest_cluster_num, &commas(&cluster_to_lba($highest_cluster_num));
 }
} # sub display_bootsect


# Read a sector from specified LBA sector number
sub read_lba_sect {
  local( $lba_offset ) = @_;
  local( $fileofs ) = (512 * $lba_offset);
  local( $fbuf, $rc );
# print "dbg0/3: seek(,$fileofs,0); sysread(,,512)\n";
  if (!seek( DISKIMAGE, $fileofs, 0 )) {
     print "ERROR: Seek(,$fileofs,0) failed, $!\n";
     &quit(2,1);
  }
  $rc = sysread(DISKIMAGE, $fbuf, 512);
  if (512 != $rc) {
     print "ERROR: Could not read '$drive' at ofs=$fileofs, $!\n";
     &quit(2,1);
  }
  return $fbuf;
} # sub read_lba_sect


# Call this to make sure that the main boot sector for the partition of
# interest has been read, that global $lba_of_bootsect has been set, and
# that &unpack_boot_sect_to_globals() and &calc_offsets have been called
# to set various global variables from the boot sector contents.
sub ensure_bootsect_loaded {
  local( $lba_to_load ) = @_;
  if (!defined $lba_to_load) {
     # no parameter specified to this function
     if (defined $lba_of_bootsect) {
        # if it was already called, then just return; we're done
        return;
     }
     # otherwise, use the command-line parameter
     $lba_to_load = $lba_offset1;
  }
  if ($lba_to_load != $lba_of_bootsect) {
     # print "dbg: ensure_bootsect_loaded: read LBA $lba_to_load.\n";
     $bootsect = &read_lba_sect( $lba_to_load );  # bootsect is a global side-effect
     $lba_of_bootsect = $lba_to_load;  # global side-effect (prevents unnecessary rereading)
     &unpack_boot_sect_to_globals( $bootsect );  # lots more global side-effects
     &calc_offsets;  # even more global side-effects
  }
} # sub ensure_bootsect_loaded


# Read a FAT32 boot sector, compare it to its backup, score each with various
# sanity checks, and report on the results.
sub check_bootsect_against_backup {
  local( $lba_offset ) = @_;
  &ensure_bootsect_loaded( $lba_offset );
  local( $fbuf1 ) = $bootsect;
  local( $score1 ) = &test4bootsect( $lba_offset );
  $b_sectorOfBackupBootSect1 = $b_sectorOfBackupBootSect;  # normally it is 6
  local( $backup_bootsec_is_best ) = 0;
  local( $lba_of_backup ) = 6 + $lba_offset;
  local( $OemId ) = $b_OemId;
  &ensure_bootsect_loaded( $lba_of_backup );
  local( $fbuf2 ) = $bootsect;
  local( $score2 ) = &test4bootsect( $lba_of_backup );
  if ($fbuf1 eq $fbuf2) {
     printf "Good!  Main and backup boot sectors are identical.  Score = %4.1f.\n", $score1;
  } elsif ('NTFS    ' eq $OemId) {
     printf "NTFS volumes don't have a backup boot sector.\n";
     $backup_bootsec_is_best = 0;
  } else {
     # here we lower the scores if the backup boot sector is bad
     if ($score1 >= 1) {
        $score1 -= 0.1;
     }
     if ($score2 >= 1) {
        $score2 -= 0.1;
     }
     if ($score2 < 7) {
        if ($score1 >= 1) {
           $score1 -= 0.5;
        }
     }
     if (6 != $b_sectorOfBackupBootSect1) {
        print "WARNING: SectorOfBackupBootSect (at 32h) should be 6, but is actually $b_sectorOfBackupBootSect1.\n";
     }
     print "WARNING: backup boot sector at LBA $lba_of_backup does not match main boot sector at LBA $lba_offset.\n";
     printf "Main boot sector score = %4.1f, backup boot sector score = %4.1f.\n", $score1, $score2;
     if ($score2 > $score1) {
        if (($score1 < 8.5) && ($score2 > 9)) {
           print "ERROR: main boot sector is damaged, but backup boot sector at LBA $lba_of_backup is good.\n";
           $backup_bootsec_is_best = 1;
        } else {
           if ($score2 < 7) {
              print "ERROR: both main and backup boot sectors are in bad shape.\n";
           } else {
              print "WARNING: backup boot sector appears to be in better shape than main boot sector.\n";
              $backup_bootsec_is_best = 1;
           }
        }
     } else {
        if ($score1 < 7) {
           print "ERROR: both main and backup boot sectors are in bad shape";
           if ($score1 > $score2) {
              print " (but main is better than backup)";
           }
           print ".\n";
        } else {
           if ($score1 < 8.5) {
              print "Note: main boot sector might be okay";
           } else {
              print "Note: main boot sector is probably okay";
           }
           if ($score1 > $score2) {
              print " (and it's better than the backup).\n";
           } else {
              print " (and it's as good as the backup).\n";
           }
        }
     }
  }
  if (! $backup_bootsec_is_best) {
     &ensure_bootsect_loaded( $lba_offset );  # restore it
  }
} # sub check_bootsect_against_backup



# Note: there are four different "bases" used by Microsoft for sector numbers.
# 1) The "LBA" (Logical Block Address) is the absolute (0-based) sector
#    number on the disk.
# 2) The "relative" sector number is the sector number relative to the partition's
#    boot sector (which is relative sector no. 0).
# 3) The most obscure base for numbering sectors is the sector number relative
#    to the partition table sector which points to this partition.  The offset
#    of the partition's boot sector relative to this base is $b_numHiddenSectors.
# 4) The implied base for the data area of the partition, which would be the
#    LBA sector number of the first data sector in cluster no. 0, if there were
#    a cluster no. 0.  (Actually, the first cluster is cluster no. 2.)


sub lba_to_rel {
  local( $lba ) = @_;
  return $lba - $lba_of_bootsect;
}


sub rel_to_lba {
  local( $rel ) = @_;
  return $rel + $lba_of_bootsect;
}


sub cluster_to_rel {
  local( $clust ) = @_;
  if (!defined $clust) {
     print "ERR: cluster_to_rel\n";
     &quit(2,1);
  }
  local( $result ) = $rel_sect_of_cluster0 + ($clust * $b_sectorsPerCluster);
  return $result;
}


sub rel_to_cluster {
  local( $rel ) = @_;
  local( $result );
  $result = int( ($rel - $rel_sect_of_cluster0) / $b_sectorsPerCluster );
  return $result;
}


sub cluster_to_lba {
  local( $clust ) = @_;
  return &rel_to_lba( &cluster_to_rel( $clust ) );
}


sub lba_to_cluster {
  local( $lba ) = @_;
  return &rel_to_cluster( &lba_to_rel( $lba ) );
}


$num_invalid_clusterrs = 0;

# Input is a cluster number; returns 1 if it is valid/reasonable, 0 if not.
sub is_valid_clusterno {
  local( $clust ) = @_;
  local( $result ) = 1;
  if ($clust < 2) {
     $result = 0;
  } elsif ($clust > $highest_cluster_num) {
     if ($o_debugmode) {
        if ($num_invalid_clusterrs < 50) {
           # the first 50 times we display a message
           printf "dbg is_valid_clustno: 'clust' %s (0x0%X) is illegal; highest_cluster_num = %s\n",
                  &commas5($clust), $clust, &commas5($highest_cluster_num);
           if ($num_invalid_clusterrs < 2) {
              # the first couple of times we also display a callback trace
              &traceback;
           }
        }
        $num_invalid_clusterrs++;
     }
     $result = 0;
  }
  return $result;
} # sub is_valid_clusterno


# Input is a FAT entry, either a cluster number or 0 or an end-chain or bad-block maker.
# Returns 1 if it is valid/reasonable, 0 if not.
#
# Note the difference between &is_valid_clusterno() and &is_valid_FAT_entry().
# &is_valid_FAT_entry() is much more lenient, it succeeds for anything that can appear in a FAT.
# &is_valid_FAT_entry() only succeeds for actual valid cluster numbers.
sub is_valid_FAT_entry {
  local( $clust ) = @_;
  local( $result ) = 1;
  if (0 == $clust) {
     $result = 1;  # zero means unallocated
  } elsif ($clust < 0) {
     $result = 0;
  } elsif ($clust >= 0x10000000) {
     $result = 0;
  } elsif ($clust == 1) {
     $result = 0;  # There is no cluster no. 1; the first legit cluster number is 2
  } elsif (($clust >= 2) && ($clust <= $highest_cluster_num)) {
     $result = 1;
  } elsif (($clust >= 0x0FFFFFF7) && ($clust <= 0x0FFFFFFF)) {
     $result = 1;  # 0x0FFFFFF8 - 0x0FFFFFFF are legit FAT entries (end-chain markers)
                   # 0x0FFFFFF7 is also a legit FAT entry (bad-block marker)
  } else {
     $result = 0;
  }
  return $result;
} # sub is_valid_FAT_entry


# simple byte-by-byte hex+ASCII dump of a some multiple of 16 bytes (e.g., a 512 byte sector)
sub dump_hex_and_ASCII {
  local( $sect ) = @_;
  local( $i, $j, $b, $c );
  local( $stop ) = ((!defined $sect) || ($sect eq ''));
  for ($i=0; (($i < length($sect)) && !$stop); $i+=16) {
     for ($j=0; $j < 16; $j++) {
        if (($i+$j) < length($sect)) {
           $c = substr( $sect, $i+$j, 1 );
           $b = ord($c);
           printf "%02X ", $b;
        } else {
           print "   ";
           $stop = 1;
        }
     }
     print " ";
     for ($j=0; $j < 16; $j++) {
        if (($i+$j) < length($sect)) {
           $c = substr( $sect, $i+$j, 1 );
           $b = ord($c);
           if (($b >= 32) && ($b <= 127)) {
              printf "%s", $c;
           } else {
              print ".";
           }
        }
     }
     print "\n";
  }
} # sub dump_hex_and_ASCII


sub dump_by_sector {
   local( $cluster ) = @_;
   do {
      $sect = substr( $cluster, 0, 512 );
      $cluster = substr( $cluster, 512 );
      &dump_hex_and_ASCII( $sect );
   } until ('' eq $cluster );
}


# simple byte-by-byte hex dump of a sector, in 4-byte groups
sub dump_sect1 {
  local( $sect ) = @_;
  local( $i, $col, $b );
  for ($i=$col=0; $i < 512; $i++) {
     $b = ord(substr( $sect, $i, 1 ));
     if ($col > 0) {
        if (0 == ($i % 4)) {
           print "  ";
        } else {
           print ".";
        }
     }
     printf "%02X", $b;
     if (++$col == 32) {
        print "\n";
        $col=0;
     }
  }
}


# dump a sector as 32-bit words (good for display of FATs)
sub dump_sect2 {
  local( $sect ) = @_;
  local( $i, $col, $b, $i2 );
  for ($i=$col=0; $i < 512; $i++) {
     $i2 = $i + 3 - (2 * ($col % 4));
     $b = ord(substr( $sect, $i2, 1 ));
     if (($col > 0) && (0 == ($col % 4))) {
        print " ";
     }
     printf "%02X", $b;
     if (++$col == 32) {
        print "\n";
        $col=0;
     }
  }
}


# Read an entire FAT at once - might be many megabytes!
# Can also be used to read other parts of the disk.
# Input is the file offset in bytes, and the number of sectors to read.
# Output is a string containing everything that was read.
sub bigread1 {
  local( $fileofs, $numsect ) = @_;
  local( $result );
  local( $amt ) = (512 * $numsect);
  if ($o_debugmode) {
     printf "dbg, bigread1: seek(,$fileofs=0x0%X,0); sysread(,,$amt)\n", $fileofs;
  }
  if (!seek( DISKIMAGE, $fileofs, 0 )) {
     print "ERROR: [1] Seek(,$fileofs,0) failed, $!\n";
     print "  from bigread1($fileofs,$numsect)\n";
     &quit(2,1);
  }
  local($rc) = sysread(DISKIMAGE, $result, $amt);
  if ($amt != $rc) {
     print "ERROR: Read failed, $amt bytes at ofs=$fileofs, rc=$rc, $!\n";
     &quit(2,1);
  }
  if ($amt != length($result)) {
     printf "ERROR: Attempt to read $amt bytes yielded %d bytes, instead.\n", length($result);
     &quit(2,1);
  }
  return $result;
} # sub bigread1


sub bigwrite1 {
   local( $fileofs, $bufr ) = @_;
   local( $amt ) = length($bufr);
   if ($o_debugmode) {
      printf "dbg, bigwrite1: seek(,$fileofs=0x0%X,0); syswrite(,,$amt)\n", $fileofs;
   }
   if (!seek( DISKIMAGE, $fileofs, 0 )) {
      print "ERROR: [1] Seek(,$fileofs,0) failed, $!\n";
      print "  from bigwrite1($fileofs, [$amt bytes] )\n";
      &quit(2,1);
   }
   local($rc) = syswrite(DISKIMAGE, $bufr, $amt);
   if (!defined $rc) {
      print "ERROR: syswrite failed, $!\n";
   } elsif ($rc != $amt) {
      print "ERROR: syswrite, requested write of $amt bytes, but only $rc were written.\n";
   }
} # sub bigwrite1


# Read main FAT into $fat, but only read it once.  (To force re-read, set $fat=''.)
sub ensure_fat_loaded {
  if ((!defined $fat) || ('' eq $fat)) {
     &ensure_bootsect_loaded;  # make sure $rel_sect_of_FAT1 is set correctly
     local( $fileofs1 ) = 512 * &rel_to_lba($rel_sect_of_FAT1);
     $fat = &bigread1( $fileofs1, $b_sectorsPerFAT );
  }
}


# Write FAT #1 to disk.
sub write_FAT1 {
   if ((!defined $fat) || ('' eq $fat)) {
      print "ERR: write_FAT1, but fat not loaded.\n";
      &quit(2,1);
   }
   local( $fileofs1 ) = 512 * &rel_to_lba($rel_sect_of_FAT1);
   &bigwrite1( $fileofs1, $fat );
}


# Read 2nd FAT into $fat2, but only read it once.  (To force re-read, set $fat2=''.)
sub ensure_fat2_loaded {
  if ($o_minmem) {
     print "ERROR: the '-m' option is incompatible with these options.\n";
     &quit;
  }
  if ((!defined $fat2) || ('' eq $fat2)) {
     &ensure_bootsect_loaded;  # make sure $rel_sect_of_FAT1 is set correctly
     local( $fileofs2 ) = 512 * &rel_to_lba($rel_sect_of_FAT2);
     $fat2 = &bigread1( $fileofs2, $b_sectorsPerFAT );
  }
}


# number of FAT sectors which differ between FAT #1 and FAT #2:
$numne = 0;
$numne_dumped = 0;


# Displays one FAT sector (if the two fats differ, then show both versions).
# Pass the sector number (relative to the beginning of the FAT), plus the
# two 512-byte sectors (the same sector, but from each copy of the FAT).
# The last parameter, $fsec2, is optional.
sub show1fatSector {
  local( $fsecno, $fsec1, $fsec2 ) = @_;
  local($both_specified) = ((defined $fsec2) && (0 != length($fsec2)));
  if ((! $both_specified) || ($fsec1 eq $fsec2)) {
     # if the sectors match in the two FATs, then display it just once
     printf "\nFAT sector no. %s (for clusters no. %s = 0x0%X thru 0x0%x)",
            &commas5($fsecno), &commas5($fsecno*128), $fsecno*128, ($fsecno*128)+127;
     if ($both_specified) {
        print " is identical in both FATs:\n";
     } else {
        print ":\n";
     }
     &dump_sect2( $fsec1 );
  } else {
     # for non-matching FAT sectors, display both versions
     printf "\nFAT sector no. %s (for clusters no. %s = 0x0%X thru 0x0%x) doesn't match in the two FATs.\n",
            &commas5($fsecno), &commas5($fsecno*128), $fsecno*128, ($fsecno*128)+127;
     printf "FAT #1 (LBA %s):\n", &commas5(&rel_to_lba($rel_sect_of_FAT1+$fsecno));
     &dump_sect2( $fsec1 );
     printf "FAT #2 (LBA %s):\n", &commas5(&rel_to_lba($rel_sect_of_FAT2+$fsecno));
     &dump_sect2( $fsec2 );
  }
} # sub show1fatSector


sub FatEqualityMsg {
  local($FATsAreUnequal) = @_;
  if ($FATsAreUnequal) {
     printf "WARNING:  The two FATs are %s sectors each, but NOT identical.\n", &commas5($b_sectorsPerFAT);
  } else {
     printf "Good!  The two FATs are %s sectors each, and identical.\n", &commas5($b_sectorsPerFAT);
  }
}


sub show_FAT_locations {
  &ensure_bootsect_loaded( $lba_offset );
  local( $fileofs1 ) = 512 * &rel_to_lba($rel_sect_of_FAT1);
  local( $fileofs2 ) = 512 * &rel_to_lba($rel_sect_of_FAT2);
  # printf "FAT #1 is at relative sector %s (LBA %s = byte offset %s)\n",
  #        &commas5($rel_sect_of_FAT1), &commas5(&rel_to_lba($rel_sect_of_FAT1)), &commas5($fileofs1);
  printf "  FAT #1 is at LBA %s = byte offset %s = relative sector %s\n",
         &commas5(&rel_to_lba($rel_sect_of_FAT1)), &commas5($fileofs1), &commas5($rel_sect_of_FAT1);
  # printf "  FAT #2 is at relative sector %s (LBA %s = byte offset %s)\n",
  #        &commas5($rel_sect_of_FAT2), &commas5(&rel_to_lba($rel_sect_of_FAT2)), &commas5($fileofs2);
  if ($rel_sect_of_FAT1 == $rel_sect_of_FAT2) {
     print "  There is no 2nd FAT.\n";
  } else {
     printf "  FAT #2 is at LBA %s = byte offset %s = relative sector %s\n",
            &commas5(&rel_to_lba($rel_sect_of_FAT2)), &commas5($fileofs2), &commas5($rel_sect_of_FAT2);
  }
  printf "  The first data sector after FAT #%d is LBA %s\n", (1 + ($rel_sect_of_FAT1 != $rel_sect_of_FAT2)),
         &commas5( $b_sectorsPerFAT + &rel_to_lba($rel_sect_of_FAT2) );
  printf "  Cluster #2 (first real cluster) is at LBA %s\n", &commas5(&cluster_to_lba(2));
  printf "       Imaginary cluster #0 would be at LBA %s\n", &commas5(&cluster_to_lba(0));
}


sub compare_two_FATs {
  local( $lba_offset ) = @_;
  local( $fsecno, $fofs, $fsec1, $fsec2 );
  $numne = $numne_dumped = 0;
  &ensure_bootsect_loaded( $lba_offset );
  &ensure_fat_loaded;  # load first FAT into $fat
  local( $fileofs1 ) = 512 * &rel_to_lba($rel_sect_of_FAT1);
  local( $fileofs2 ) = 512 * &rel_to_lba($rel_sect_of_FAT2);
  local( $FATsAreUnequal ) = 0;
  local( $sectrsAreEqual ) = 0;
  local( $showit ) = 0;
  if ($o_minmem) {
     # don't load entire 2nd FAT into memory
     for ($fsecno=0; $fsecno<$b_sectorsPerFAT; $fsecno++) {
        $fofs = 512 * $fsecno;
        $fsec1 = substr( $fat, $fofs, 512 );
        # $fsec1 = &read_lba_sect( &rel_to_lba($rel_sect_of_FAT1+$fsecno) );  # temporary
        $fsec2 = &read_lba_sect( &rel_to_lba($rel_sect_of_FAT2+$fsecno) );
        if ($fsec1 eq $fsec2) {
           $sectrsAreEqual = 1;
        } else {
           $sectrsAreEqual = 0;
           $numne++;
        }
        $showit = ($fsecno < $o_showFAT) || ((!$sectrsAreEqual) && ($o_debugmode || $o_verbose) && ($numne <= 20));
        if ($showit) {
           &show1fatSector( $fsecno, $fsec1, $fsec2, $o_showFAT );
           $numne_dumped++;
        }
     }#for
     $FATsAreUnequal = (0 != $numne);
     &FatEqualityMsg( $FATsAreUnequal );
  } else {
     # normal mode -- load both FATs into memory, all at once
     &ensure_fat2_loaded;  # load 2nd FAT into $fat2
     $FATsAreUnequal = ($fat ne $fat2);
     &FatEqualityMsg( $FATsAreUnequal );
     if ($FATsAreUnequal || $o_debugmode || $o_verbose || $o_showFAT) {
        # The two FATs are not identical.  So compare 'em sector-by-sector:
        for ($fsecno=0; $fsecno<$b_sectorsPerFAT; $fsecno++) {
           $fofs = 512 * $fsecno;
           $fsec1 = substr( $fat, $fofs, 512 );
           $fsec2 = substr( $fat2, $fofs, 512 );
           if ($fsec1 eq $fsec2) {
              $sectrsAreEqual = 1;
           } else {
              $sectrsAreEqual = 0;
              $numne++;
           }
           $showit = ($fsecno < $o_showFAT) || ((!$sectrsAreEqual) && ($o_debugmode || $o_verbose) && ($numne <= 20));
           if ($showit) {
              &show1fatSector( $fsecno, $fsec1, $fsec2 );
              $numne_dumped++;
           }
        }#for
     }
  }
  if ($FATsAreUnequal) {
     printf "%s of %s sectors match, %s don't.\n",
            &commas5($b_sectorsPerFAT-$numne), &commas5($b_sectorsPerFAT), &commas5($numne);
     if (($numne_dumped > 0) && ($numne > $numne_dumped)) {
        printf "(%s mismatches not dumped)\n", &commas5($numne-$numne_dumped);
     }
  }
} # sub compare_two_FATs


# Pass in either the maximum number of FAT sectors to display, or the
# first and last sector numbers (relative to start of FAT).
sub dbg_FAT_dmp {
  local( $firstsect, $lastsect ) = @_;
  if (0 == length($fat)) {
     print "dbg_FAT_dmp: fat not loaded.\n";
     return;
  }
  if (0 == length($fat2)) {
     print "dbg_FAT_dmp: fat2 not loaded.\n";
     return;
  }
  if (! defined $lastsect) {
     if (! defined $firstsect) {
        $firstsect = 1;
     }
     $lastsect = $firstsect-1;
     $firstsect = 0;
  }
  if ($firstsect < 0) {
     $firstsect = 0;
  }
  if ($lastsect < 0) {
     $lastsect = 0;
  }
  if ($lastsect >= $b_sectorsPerFAT) {
     $lastsect = ($b_sectorsPerFAT - 1);
  }
  if ($firstsect > $lastsect) {
     $firstsect = $lastsect;
  }
  print "dbg_FAT_dmp(,$firstsect,$lastsect)...\n";
  local( $fsecno, $fofs, $fsec1, $fsec2 );
  &show_fat_locations;
  for ($fsecno = $firstsect; $fsecno <= $lastsect; $fsecno++) {
     $fofs = 512 * $fsecno;
     $fsec1 = substr( $fat, $fofs, 512 );
     $fsec2 = substr( $fat2, $fofs, 512 );
     &show1fatSector( $fsecno, $fsec1, $fsec2 );
  }#for
} # sub dbg_FAT_dmp


# for Perl4 only
# sub chr {
#    local($x) = @_;
#    local($result) = pack("c",$x);
#    return $result;
# }


# Rate a 32-bit FAT entry according to how likely it seems to be correct.
# If it is an invalid entry (e.g., 0xFFFFFFFF) then rate it -50.
# If it is zero (indicating an unused cluster) then rate it 0.
# If it is a valid end-of-chain entry, 0x0FFFFFFF, then rate it 3.
# if it is a valid bad-block indicator, 0x0FFFFFF7, then rate it 0.
# If it points to the next contiguous entry, and that entry is valid & non-zero, then rate it 10.
# If it points to the next contiguous entry, but that entry is invalid or zero, then rate it 2.
# If it points to a later (discontiguous) entry, and that entry is valid & non-zero, then rate it 2.
# If it points to a later (discontiguous) entry, and that entry is invalid or zero, then rate it 0.
# If it points to an earlier entry, and that entry is valid & non-zero, then rate it 1.
# If it points to an earlier entry, and that entry is invalid or zero, then rate it -1.
# Currently, I make no effort to check the FAT entries for consistency with the directories.
# I will also use this function to rate entire sectors for quality, simply by adding the
# ratings for all 128 entries together.
# Input parameters are:
#   $fsecno -- sector number of the FAT sector containing this FAT entry; first FAT sector is #0.
#   $fsectindex -- index into the FAT sector of this FAT entry, 0..127.
#   $entry -- the 32-bit FAT entry to be rated.
#   $which_fat -- 1 or 2.
sub rate_fat_entry {
   local( $fsecno, $fsectindex, $entry, $which_fat ) = @_;
   local( $result ) = 0;
   local( $fentryno ) = (128 * $fsecno) + $fsectindex;
   local( $nxtcontig ) = $fentryno + 1;
   if (($which_fat < 1) || ($which_fat > 2)) {
      print "ERR: rate_fat_entry\n";
      &quit(2,1);
   }
   if ((! &is_valid_FAT_entry($entry)) || ($entry == $fentryno)) {
      $result = -50;
   } elsif (0 == $entry) {
      $result = 0;
   } elsif (0x0FFFFFFF == $entry) {
      $result = 3;
   } elsif (0x0FFFFFF7 == $entry) {
      $result = 0;
   } else {
       local( $pointsto ) = &FatEntry( $entry, $which_fat );
       if ($entry == $nxtcontig) {
          if ($pointsto > 0) {
             $result = 10;
          } else {
             $result = 2;
          }
       } elsif ($entry > $nxtcontig) {
          if ($pointsto > 0) {
             $result = 2;
          } else {
             $result = 0;
          }
       } elsif ($entry < $fentryno) {
          if ($pointsto > 0) {
             $result = 1;
          } else {
             $result = -1;
          }
       } else {
          print "ERR: rate_fat_entry/2\n";
          $result = -1;
       }
   }
   # if ($o_debugmode > 1) {
   #    print "dbg: rate_fat_entry($fsecno,$fsectindex," . &DispFatEnt($entry) . ",$which_fat) = $result\n";
   # }
   return $result;
} # sub rate_fat_entry


sub rate_fat_sector {
   local( $fsecno, $sect, $which_fat ) = @_;
   local( $fsectindex );
   local( $result ) = 0;
   local( $firstindex ) = 0;
   if (0 == $fsecno) {
      # in sector #0 we skip the first two entries, since the first cluster is #2.
      $firstindex = 2;
   }
   for ($fsectindex=$firstindex; $fsectindex<=127; $fsectindex++) {
      $tmp = substr( $sect, 4 * $fsectindex, 4 );
      ($tmp) = unpack( "L", $tmp );
      $result += &rate_fat_entry( $fsecno, $fsectindex, $tmp, $which_fat );
   }
   if ($o_debugmode >= 3) {
      print "dbg: rate_fat_sector($fsecno,,$which_fat) = $result\n";
   }
   return $result;
} # sub rate_fat_sector


sub poke_sector_in_fat {
   local( $afat, $fsectno, $newsect ) = @_;
   local( $fofs ) = 512 * $fsectno;
   $afat = substr( $afat, 0, $fofs ) . $newsect . substr( $afat, $fofs+512 );
   return $afat;
}


sub poke_sector_in_afat {
   local( $whichfat, $fsectno, $newsect ) = @_;
   local( $fofs ) = 512 * $fsectno;
   local($afat);
   if (2 == $whichfat) {
      $afat = substr( $fat2, 0, $fofs ) . $newsect . substr( $fat2, $fofs+512 );
   } else {
      $afat = substr( $fat, 0, $fofs ) . $newsect . substr( $fat, $fofs+512 );
   }
   return $afat;
}


$timer{'poke_longword_in_afat'} = 0.0;

# inverse of &FatEntry
# needed to implement '-c' ($o_correctFATfromDirs)
sub poke_longword_in_afat {
   local( $whichfat, $clusterno, $newentry ) = @_;
   local( $fofs ) = 4 * $clusterno;
   local($afat);
   $newentry = pack("L",$newentry);
   if (2 == $whichfat) {
      $afat = substr( $fat2, 0, $fofs ) . $newentry . substr( $fat2, $fofs+4 );
   } else {
      $afat = substr( $fat, 0, $fofs ) . $newentry . substr( $fat, $fofs+4 );
   }
   return $afat;
}



$timer{'reconcile_two_FATs'} = 0.0;
$timer{'reconcile_two_FATs_sectorpoking'} = 0.0;

# When one or both of the FATs are damaged, by having some of the FAT sectors
# overwritten with garbage, we can often fix them simply by comparing the two
# FATs sector-by-sector, and choosing the best sector from each FAT.  This is
# the function which does that, to implement the '-r' and '-R' options.
# Note: it is slow!
sub reconcile_two_FATs {
   local( $lba_offset ) = @_;
   local( $fsecno, $fofs, $fsec1, $fsec2 );
   $numne = 0;
   local( $numSectStillBad ) = 0;
   &ensure_bootsect_loaded( $lba_offset );
   &ensure_fat_loaded;  # load first FAT into $fat
   local( $fileofs1 ) = 512 * &rel_to_lba($rel_sect_of_FAT1);
   local( $fileofs2 ) = 512 * &rel_to_lba($rel_sect_of_FAT2);
   printf "Reconciling the two FATs...\n";

   # printf "  FAT #1 is at relative sector %s (LBA %s = byte offset %s)\n",
   #        &commas5($rel_sect_of_FAT1), &commas5(&rel_to_lba($rel_sect_of_FAT1)), &commas5($fileofs1);
   # printf "  FAT #2 is at relative sector %s (LBA %s = byte offset %s)\n",
   #        &commas5($rel_sect_of_FAT2), &commas5(&rel_to_lba($rel_sect_of_FAT2)), &commas5($fileofs2);
   # printf "  Cluster #0 is at LBA %s\n", &commas5(&cluster_to_lba(0));

   # &show_fat_locations;

   &ensure_fat2_loaded;  # load 2nd FAT into $fat2
   if ($fat eq $fat2) {
      printf "The two FATs are %s sectors each, and identical, so reconcilliation isn't possible.\n", &commas5($b_sectorsPerFAT);
      return;
   }
   $timer{'reconcile_two_FATs'} -= (times)[0];

   local($fatlen) = $b_sectorsPerFAT * 512;
   if ($fatlen != length($fat)) {
      print "ERROR: calculated FAT length <> amt read from FAT #1.\n";
   }
   if ($fatlen != length($fat2)) {
      print "ERROR: calculated FAT length <> amt read from FAT #2.\n";
   }

   if ($o_debugmode > 1) {
      print "before reconciliation, "; &dbg_FAT_dmp(0,7);
   }

   # The two FATs are not identical.  So compare 'em sector-by-sector:
   local($best);
   $fat1fixes = $fat2fixes = 0;
   local($rating1) = local($rating2) = 0;
   local($NotIdenticalButRatedSame) = 0;
   for ($fsecno=0; $fsecno<$b_sectorsPerFAT; $fsecno++) {
      $fofs = 512 * $fsecno;
      $fsec1 = substr( $fat, $fofs, 512 );
      $fsec2 = substr( $fat2, $fofs, 512 );
      $rating1 = &rate_fat_sector( $fsecno, $fsec1, 1 );
      $best = "";
      if ($fsec1 eq $fsec2) {
         $rating2 = $rating1;
         if ($rating1 < 0) {
            if ($o_debugmode) {
               printf "Warning: FAT sector no. %s is identical in the 2 FATs & corrupted in both (rating $rating1)\n", &commas5($fsecno);
               &show1fatSector( $fsecno, $fsec1, $fsec2 );
            }
         }
      } else {
         $numne++;
         $rating2 = &rate_fat_sector( $fsecno, $fsec2, 2 );
         if ($o_debugmode) {
            printf "FAT sector no. %s differs in the two FATs (ratings are $rating1 \& $rating2):\n", &commas5($fsecno);
            &show1fatSector( $fsecno, $fsec1, $fsec2 );
         }
         if ($rating1 > $rating2) {
            $best = $fsec1;
         } elsif ($rating2 > $rating1) {
            $best = $fsec2;
         } else {
            # Note: if the two ratings are equal, we leave 'em alone.
            $NotIdenticalButRatedSame++;
         }

         # At this point we could consider comparing entry-by-entry instead
         # of sector-by-sector... for some kinds of damage, that might work
         # better.
      }
      if ("" ne $best) {
         $timer{'reconcile_two_FATs_sectorpoking'} -= (times)[0];
         if ($best ne $fsec1) {
            $fat1fixes++;
            # $fat = substr( $fat, 0, $fofs ) . $best . substr( $fat, $fofs+512 );
            # $fat = &poke_sector_in_fat( $fat, $fsecno, $best );  # -- this uses less memory but takes more time
            $fat = &poke_sector_in_afat( 1, $fsecno, $best );  # -- this takes less time but uses more memory
            if ($o_debugmode) {
               printf "Copying sector $fsecno from FAT#2 (rated $rating2) to FAT#1 (rated $rating1).\n";
            }
         }
         if ($best ne $fsec2) {
            $fat2fixes++;
            #$fat2 = substr( $fat2, 0, $fofs ) . $best . substr( $fat2, $fofs+512 );
            # $fat2 = &poke_sector_in_fat( $fat2, $fsecno, $best );
            $fat2 = &poke_sector_in_afat( 2, $fsecno, $best );
            if ($o_debugmode) {
               printf "Copying sector $fsecno from FAT#1 (rated $rating1) to FAT#2 (rated $rating2).\n";
            }
         }
         $timer{'reconcile_two_FATs_sectorpoking'} += (times)[0];
      }
      if (($rating1 < 0) && ($rating2 < 0)) {
         $numSectStillBad++;
         if ($o_debugmode) {
            print "Warning: BOTH VERSIONS OF FAT SECTOR $fsecno LOOK BAD!\n";
         }
      }
   }#for
   printf "%s of %s sectors match already.\n", &commas5($b_sectorsPerFAT-$numne), &commas5($b_sectorsPerFAT);
   print "fixed $fat1fixes sectors in FAT #1 (by copying them from FAT #2).\n";
   print "fixed $fat2fixes sectors in FAT #2 (by copying them from FAT #1).\n";
   if ($NotIdenticalButRatedSame) {
      print "$NotIdenticalButRatedSame sectors are not identical, but have equal ratings so were left alone.\n";
   }
   print "$numSectStillBad FAT sectors are still corrupt after reconciliation.\n";
   if ($fatlen != length($fat)) {
      print "ERR: calculated FAT length <> size of FAT #1 after reconciliation.\n";
   }
   if ($fatlen != length($fat2)) {
      print "ERR: calculated FAT length <> size of FAT #2 after reconciliation.\n";
   }
   $timer{'reconcile_two_FATs'} += (times)[0];
   printf "Reconciliation complete in %.2f secs (%0.2f secs spent poking sectors).\n  ",
          $timer{'reconcile_two_FATs'}, $timer{'reconcile_two_FATs_sectorpoking'};
   &get_stat_info;
   if ($o_debugmode > 1) {
      print "after reconciliation, "; &dbg_FAT_dmp(0,7);
   }
} # sub reconcile_two_FATs


sub dump_FAT1 {
  local( $lba_offset ) = @_;  # LBA offset of boot sector
  local( $fsecno, $fofs, $fsec1 );
  &ensure_bootsect_loaded( $lba_offset );
  &ensure_fat_loaded;  # load FAT into $fat
  local( $fileofs1 ) = 512 * &rel_to_lba($rel_sect_of_FAT1);
  printf "FAT is at relative sector $rel_sect_of_FAT1 (LBA " . &commas5(&rel_to_lba($rel_sect_of_FAT1)) . " = byte offset %s)\n", &commas5($fileofs1);
  for ($fsecno=0; $fsecno<$b_sectorsPerFAT; $fsecno++) {
     $fofs = 512 * $fsecno;
     $fsec1 = substr( $fat, $fofs, 512 );
     # $fsec1 = &read_lba_sect( &rel_to_lba($rel_sect_of_FAT1+$fsecno) );  # temporary
     if ($fsecno < $o_showFAT) {
        # just dump the first $o_showFAT (perhaps 100) FAT sectors
        &show1fatSector( $fsecno, $fsec1 );
     }
  }#for
} # sub dump_FAT1


# examine the contents of $fat, and display some statistics, such as % disk usage
sub examine_FAT {
  local( $lba_offset ) = @_;
  local( $zero_sectors, $fsecno, $fofs, $fsec1, $bofs, $entry, $ientry, $used, $invalid, $num_sectrs );
  $zero_sectors = $used = $invalid = $chainends = $badblocks = $ientry = 0;
  &ensure_bootsect_loaded( $lba_offset );
  &ensure_fat_loaded;  # load first FAT into $fat
  local( $fileofs1 ) = 512 * &rel_to_lba($rel_sect_of_FAT1);
  # local( $fileofs2 ) = 512 * &rel_to_lba($rel_sect_of_FAT2);
  # print "FAT #1 is at relative sector $rel_sect_of_FAT1 (LBA " . &rel_to_lba($rel_sect_of_FAT1) . ", byte offset $fileofs1)\n";
  # print "FAT #2 is at relative sector $rel_sect_of_FAT2 (LBA " . &rel_to_lba($rel_sect_of_FAT2) . ", byte offset $fileofs2)\n";
  $num_sectrs = $b_sectorsPerFAT;
  if (length($fat) != (512 * $b_sectorsPerFAT)) {
     $num_sectrs = int( length($fat) / 512 );
     printf "ERROR: FAT was incompletely read.  It should be $b_sectorsPerFAT sectors, but is actually only %.2f sectors.\n", length($fat)/512;
  }
  local($first_bofs) = 8;  # we ignore the 1st 8 bytes of the first FAT sector
  for ($fsecno=0; $fsecno<$num_sectrs; $fsecno++) {
     $fofs = 512 * $fsecno;
     $fsec1 = substr( $fat, $fofs, 512 );
     if ($fsec1 eq $zerosect) {
        $zero_sectors++;
     } else {
        for ($bofs=$first_bofs; $bofs<512; $bofs+=4) {
           $entry = substr( $fsec1, $bofs, 4 );
           if ($entry ne $four0bytes) {
              $used++;
              $ientry = unpack( "L", $entry );  # here we assume that a "signed long value" is 32-bits -- not true for all Perls
              if (0x0FFFFFFF == $ientry) {
                 $chainends++;
              } elsif (0x0FFFFFF7 == $ientry) {
                 $badblocks++;
              } elsif (! &is_valid_FAT_entry($ientry)) {
                 $invalid++;
              }
           }
        }#for
     }
     $first_bofs = 0;  # we ignore the 1st 8 bytes of the first FAT sector, but not subsequent sectors
  }#for
  printf "FAT #1 contains $zero_sectors all-zero sectors, and %d not-all-zero sectors.\n",
         ($num_sectrs - $zero_sectors);
  printf "It represents %s used clusters, which is %4.2f%% of the partition.\n",
         &commas($used), 100.0 * ($used / (128 * $num_sectrs));
  printf "It contains %s end-chains (so there are %s files \& folders in the partition)\n", &commas5($chainends), &commas5($chainends);
  if ($badblocks) {
     print "Warning: FAT #1 contains $badblocks clusters marked as BAD (0x0FFFFFF7).\n";
  }
  if ($invalid) {
     print "Warning: FAT #1 contains $invalid invalid entries.\n";
  }
} # sub examine_FAT


# Generate a ddrescue-style log file (sector list file) representing the
# clusters which contain the bitmap (to be used like 'important.log' from
# nficruncher.pl is used):
#
# A logfile looks something like this:
#
# -----------------------------------------
# # Rescue Logfile. Created by GNU ddrescue version 1.9-pre1
# # current_pos  current_status
# 0xA05550800     ?
# #      pos        size  status
# 0x00000000  0x786DD8000  +
# 0x786DD8000  0x00000200  -
# 0x786DD8200  0x761AD800  ?
# -----------------------------------------


# Format a ddrescue logfile line.  Input parameters are in sectors (512 bytes),
# but the result is displayed in bytes.
sub displayrange {
   local( $ofs, $len, $status ) = @_;
   if (!defined $status) {
      $status = 'undef';  # should never happen
   }
   local( $result ) = sprintf( "0x%06X00  0x%06X00  %s", (2*$ofs), (2*$len), $status );
   return $result;
}

# ddrescue logfile header:
$logfile_header = "# Rescue Logfile. Created by GNU ddrescue version 1.9-burtonhack7\n" .
                  "# current_pos  current_status\n" .
                  "0x0000000000     ?\n" .
                  "#      pos        size  status\n";


# create a ddrescue-style logfile (sector list) for the location OF the first FAT:
sub gen_logfile_for_FAT1_locations {
   open( LOG_OUTPUT, ">LocationOfFAT1.log" ) || die "ERROR: could not write 'LocationOfFAT1.log', $!\n";
   print LOG_OUTPUT $logfile_header;
   local($ofs) = &rel_to_lba($rel_sect_of_FAT1);
   print LOG_OUTPUT &displayrange($ofs,$b_sectorsPerFAT,'+') . "\n";
   close LOG_OUTPUT;
   print "Created 'LocationOfFAT1.log'.\n";
}



# possible values for $counting_what
$recovery_needed_1 = 1;
$recovery_not_needed_0 = 0;
# globals used by &export_FAT and its subroutines:
$counting_what = $recovery_not_needed_0;
$ofsS = 0;
$lenS = 0;


# Subroutine of &export_FAT, writes one line of a ddrescue logfile, according
# to globals $counting_what, $lenS, and $ofsS.
sub ef_writelogrec {
   if ($lenS > 0) {
      if ($counting_what == $recovery_needed_1) {
         # end of a series of sectors needing recovery
         print LOG_OUTPUT &displayrange($ofsS,$lenS,'?') . "\n";
      } else {
         # end of a series of sectors not needing recovery
         print LOG_OUTPUT &displayrange($ofsS,$lenS,'+') . "\n";
      }
      $ofsS += $lenS;
      $lenS = 0;
   }
}


# Subroutine of &export_FAT, called for a sector that still needs to be recovered.
# Note that the one parameter is optional, but if you omit it then you must
# specify an empty argument list
sub ef_sec_needs_recovery {
   local( $numsec, $sec ) = @_;
   if (!defined $numsec) {
      $numsec = 1;
   }
   if ((defined $sec) && ($sec != ($ofsS+$lenS))) {
      printf "ERR: ef_sec_needs_recovery( $numsec, $sec ) -- sector should be%d\n", $ofsS+$lenS;
      &quit(2,1);
   }
   if ($counting_what == $recovery_not_needed_0) {
      # after a series of sectors not needing recovery, a sector which does need recovery
      &ef_writelogrec;
      $counting_what = $recovery_needed_1;
   }
   $lenS += $numsec;
}

# equivalent to ef_sec_needs_recovery(1) wtih no error checks
sub ef_sec_needs_recovery_quick {
   if ($counting_what == $recovery_not_needed_0) {
      # after a series of sectors not needing recovery, a sector which does need recovery
      &ef_writelogrec;
      $counting_what = $recovery_needed_1;
   }
   $lenS++;
}


# Subroutine of &export_FAT, called for one or more sectors that needn't be
# recovered (either because they have already been recovered, or because they
# are in the unused space).
sub ef_sec_doesnt_need_recovery {
   local( $numsec, $sec ) = @_;
   if (!defined $numsec) {
      $numsec = 1;
   }
   if ((defined $sec) && ($sec != ($ofsS+$lenS))) {
      printf "ERR: ef_sec_doesnt_need_recovery( $numsec, $sec ) -- sector should be%d\n", $ofsS+$lenS;
      &quit(2,1);
   }
   if ($counting_what == $recovery_needed_1) {
      # after a series of sectors needing recovery, a sector which doesn't need recovery
      &ef_writelogrec;
      $counting_what = $recovery_not_needed_0;
   }
   $lenS += $numsec;
}

# equivalent to &ef_sec_doesnt_need_recovery(1) with no error checks
sub ef_sec_doesnt_need_recovery_quick {
   if ($counting_what == $recovery_needed_1) {
      # after a series of sectors needing recovery, a sector which doesn't need recovery
      &ef_writelogrec;
      $counting_what = $recovery_not_needed_0;
   }
   $lenS++;
}


# Export freespace list from FAT #1 in ddrescue logfile format (implement the '-e' option).
# Creates a ddrescue logfile ("fat1freespace.log") which contains status="+" lines for all
# the freespace listed in FAT #1.  By merging fat1freespace.log with the ddrescue logfile
# for a partially recovered drive, we mark the freespace as already recovered, without
# having to actually recover it.
# If -e2 is specified ($o_exportFAT2log==2) then the only sectors NOT marked with status="+"
# are those sectors which are "used" in the FAT and which are not equal to 512 bytes of zeros,
# plus sectors in the FATs themselves which are all zeros.
# This is for recovering a lost ddrescue log file in the case when the FAT is believed to
# be (at least mostly) intact.
sub export_FAT {
   local( $lba_offset ) = @_;
   &ensure_bootsect_loaded( $lba_offset );
   &ensure_fat_loaded;  # load first FAT into $fat

   &gen_logfile_for_FAT1_locations;  # create LocationOfFAT1.log

   open( LOG_OUTPUT, ">fat1freespace.log" ) || die "ERROR: could not write 'fat1freespace.log', $!\n";
   print LOG_OUTPUT $logfile_header;
   local( $prev_ofs_plus_len, $start_of_unused_range, $entry, $sec, $used, $unused, $cluster );

   if ($o_exportFAT2log >= 2) {
      print ",";  # progress indicator
      # First, export as "+" everything prior to the two FATs:
      $ofsS = 0;
      $lenS = &rel_to_lba($rel_sect_of_FAT1);
      print LOG_OUTPUT &displayrange($ofsS,$lenS,'+') . "\n";
      $ofsS += $lenS;
      $lenS = 0;
      # Next, export as "+" all non-zero FAT sectors:
      &ensure_fat2_loaded;  # load second FAT into $fat2
      print ",";  # progress indicator
      for ($sec=0; $sec < $b_sectorsPerFAT; $sec++) {
         $sectr = substr( $fat, 512*$sec, 512 );
         if ($sectr eq $zerosect) {
            &ef_sec_needs_recovery( 1, $sec + &rel_to_lba($rel_sect_of_FAT1) );
         } else {
            &ef_sec_doesnt_need_recovery( 1, ($sec + &rel_to_lba($rel_sect_of_FAT1)) );
         }
      }#for
      print ",";  # progress indicator
      for ($sec=0; $sec < $b_sectorsPerFAT; $sec++) {
         $sectr = substr( $fat2, 512*$sec, 512 );
         if ($sectr eq $zerosect) {
            &ef_sec_needs_recovery( 1, $sec + &rel_to_lba($rel_sect_of_FAT2) );
         } else {
            &ef_sec_doesnt_need_recovery( 1, ($sec + &rel_to_lba($rel_sect_of_FAT2)) );
         }
      }#for
      &ef_writelogrec;
      if ($ofsS != &cluster_to_lba(2)) {
         printf "ERR: export_FAT, ofs after FATs is %s but should be LBA %s = lba of cluster 2.\n", &commas5($ofsS), &commas5(&cluster_to_lba(2));
         &quit(2);
      }
      print LOG_OUTPUT "# End FATs, begin data clusters\n";
      print ";";  # progress indicator
   } else {
      $ofsS = &cluster_to_lba(2);  # the first cluster is cluster #2
      $lenS = 0;
   }

   $prev_ofs_plus_len = $start_of_unused_range = 0;
   $counting_what = $recovery_needed_1;
   $used = 2;  # clusters 0 & 1 are always used, so we start looking at the fat at cluster 2
   $unused = 0;
   if (&cluster_to_lba(2) != ($ofsS+$lenS)) {
      printf "ERR: export_FAT pt2, ofs+len %d rather than LBA of cluster #2 (%d)\n", $ofsS+$lenS, &cluster_to_lba(2);
      &quit(2,1);
   }
   local($seventyfifthdone) = int($highest_cluster_num / 74);  # each progress indicator dot represents about 1.33% done
   local($clust,$lbaOfClust);
   for ($clust=2; $clust <= $highest_cluster_num; $clust++) {
      if (($clust % $seventyfifthdone) == 0) {
         print ".";  # progress indicator
      }
      $entry = substr( $fat, 4*$clust, 4 );
      $lbaOfClust = &cluster_to_lba($clust);
      if (($entry eq $four0bytes) || ($entry eq $badclustermark)) {
         # unused clusters (and clusters marked as bad in the FAT) don't need recovery
         $unused++;
         &ef_sec_doesnt_need_recovery( $b_sectorsPerCluster, $lbaOfClust );
      } else {
         # used clusters might need recovery
         $used++;
         if ($o_exportFAT2log < 3) {
            # For -e2 option, we report all the used clusters as needing recovery
            &ef_sec_needs_recovery( $b_sectorsPerCluster, $lbaOfClust );
         } else {
            # For -e3 option, we read the used clusters from the drive, and only
            # report the sectors as needing recovery if they are all zeros.
            # (This is for reconstruction of a lost ddrescue logfile.)
            $cluster = &bigread1( $lbaOfClust * 512, $b_sectorsPerCluster );
            for ($sec=0; $sec < $b_sectorsPerCluster; $sec++) {
               $sectr = substr( $cluster, 512*$sec, 512 );
               if ($sectr eq $zerosect) {
                  &ef_sec_needs_recovery_quick;  # &ef_sec_needs_recovery( 1, $lbaOfClust+$sec );
               } else {
                  &ef_sec_doesnt_need_recovery_quick;  # &ef_sec_doesnt_need_recovery( 1, $lbaOfClust+$sec );
               }
            }
         }
      }
   }#for
   &ef_writelogrec;  # write the last line to the ddrescue logfile
   close LOG_OUTPUT;
   print "\n";  # end of progress indicator
   print "Created 'fat1freespace.log'.\n";
   printf "Fat indicates %s used clusters and %s unused clusters.\n", &commas5($used), &commas5($unused);
   printf "(That's %s used sectors sectors and %s unused sectors.)\n", &commas5($b_sectorsPerCluster * $used), &commas5($b_sectorsPerCluster * $unused);
} # sub export_FAT


# Input is a boot sector.  Examine it to see how much it looks like a real boot
# sector, perhaps display its contents, and set a slew of global side-effects.
# Also optionally examines and compares the two FATs, optionally reconciles
# differences between them, and optionally exports the FAT in ddrescue format.
sub check_and_display_bootsect {
  local( $lba_offset, $score ) = @_;
  if ($o_debugmode || $o_verbose || $o_showbootsect) {
     print "============================================================\n";
  }
  &ensure_bootsect_loaded( $lba_offset );
  if ($o_debugmode) {
     &dump_hex_and_ASCII( $bootsect );
  }
  if ($o_debugmode || $o_verbose || $o_showbootsect) {
     printf "\nBOOTSECTOR-ness score = %4.1f\n", $score;
     &display_bootsect( $lba_offset );
  }
  &check_bootsect_against_backup( $lba_offset );

  &show_FAT_locations;

  if (!$o_search4bootsect) {
     if ($o_reconcile || $o_reconcilewrite) {
        &reconcile_two_FATs;
     }
     if ($rel_sect_of_FAT1 == $rel_sect_of_FAT2) {
        if (0 == $forced_FAT_start_sect) {
           print "Note: there is no 2nd FAT on this file system (b_numFATs=$b_numFATs).\n";
        } else {
           &dump_FAT1( $lba_offset );
        }
     } else {
        &compare_two_FATs( $lba_offset );
     }
     print "Analyzing FAT...\n";
     &examine_FAT( $lba_offset );
     if ($o_exportFAT2log) {
        print "Exporting to ddrescue log file...\n";
        &export_FAT( $lba_offset );
     }
  }

} # sub check_and_display_bootsect


# Input is the LBA sector number of a boot sector.  Read the boot sector,
# set a slew of global side-effects, examine it to see how much it looks
# like a real boot sector, and display its contents.
sub read_check_and_display_bootsect {
   local( $lba_offset ) = @_;
   &ensure_bootsect_loaded( $lba_offset );  # sets $lba_of_bootsect as global side-effect
   local($score) = &test4bootsect( $lba_offset );
   &check_and_display_bootsect( $lba_of_bootsect, $score );  # sets lottsa global side-effects
} # sub read_check_and_display_bootsect


# Input is a cluster number.  This function reads the FAT entry for that
# cluster number, and does almost no checking for validity of input (so,
# for example, you can use it to fetch the dirty bits).
sub UncheckedFatEntry {
  local( $clus ) = @_;
  local( $result ) = -1;
  &ensure_fat_loaded;  # load first FAT into $fat
  if ((defined $fat) && (length($fat) > (4*$clus))) {
     $result = substr( $fat, 4 * $clus, 4 );
     ($result) = unpack( "L", $result );
  }
  return $result;
}


# Same, but from fat #2
sub UncheckedFat2Entry {
  local( $clus ) = @_;
  local( $result ) = -1;
  &ensure_fat2_loaded;  # load 2nd FAT into $fat2
  if ((defined $fat2) && (length($fat2) > (4*$clus))) {
     $result = substr( $fat2, 4 * $clus, 4 );
     ($result) = unpack( "L", $result );
  }
  return $result;
}



# Input is a cluster number.  This function reads the FAT entry for that
# cluster number, which is:
#   0          - means cluster is unallocated
#   nnn        - next cluster number in chain (if nnn >= 2)
#   0xFFFFFFF_ - cluster is last in chain
#   -1         - input was not a valid cluster number
# Optional 2nd input parameter is the fat number, 1 or 2.  If omitted
# or 1 then the entry is fetched from FAT #1, if 2 then it is fetched
# from FAT #2.
# Inverse of this function is &poke_longword_in_afat.
sub FatEntry {
  local( $clus, $fatnum ) = @_;
  local( $result ) = -1;
  if (!defined $fatnum) {
     $fatnum = 1;
  }
  if (&is_valid_clusterno($clus)) {
     if (2 == $fatnum) {
        $result = &UncheckedFat2Entry( $clus );
     } else {
        $result = &UncheckedFatEntry( $clus );
     }
  }
  return $result;
} # sub FatEntry


# equivalent to &FatEntry($clus,1), but slightly faster
sub Fat1Entry {
  local( $clus ) = @_;
  local( $result ) = -1;
  if (&is_valid_clusterno($clus)) {
     $result = &UncheckedFatEntry( $clus );
  }
  return $result;
} # sub Fat1Entry


# input is a cluster number
# result is an 8-character string
sub DispFatEnt {
  local( $clus ) = @_;
  local( $result );
  if ($clus > 0) {
     $result = sprintf( "%08X", $clus );
  } else {
     $result = sprintf( "%8d", $clus );  # probably '      -1'
  }
  return $result;
}


# Given a cluster number, return the cluster number of the last cluster in
# its cluster list.  This cluster number is how we'll look up a file from any
# of its sector numbers:
# first, compute the containing cluster number from the sector number;
# then find the last cluster number for the file (using this subroutine);
# then look up the file number from its last cluster number;
# then look up the file name, directory, etc. from its file number.
# Side effect results are:
#   $hops2last -- one less than number of clusters seen in chain
#   @hoplist -- list of clusters seen in chain
#   $cluster_chain_missing -- normally zero, but 1 iff the specified cluster
#        is not in use (an error condition), or N if the N-th cluster in the
#        chain is not in use (also an error condition).  If non-zero, it means
#        the cluster chain did not end properly with 0x0FFFFFFx as it should.
# Note that $hops2last = $#hoplist
sub LastClust {
   local( $frst ) = local( $nxt ) = @_;
   local( $clus, $tmp );
   # side-effect results:
   $hops2last = -1;
   $cluster_chain_missing = 0;
   @hoplist = ();
   if (! &is_valid_clusterno($nxt)) {
      printf "dbg-ERR: LastClust(0x0%X) bad-start\n", $frst;
      &traceback;
      $clus = 0;
   } else {
      $tmp = &UncheckedFatEntry( $nxt );
      if ((0 == $tmp) || (0x0FFFFFF7 == $tmp)) {
         $clus = 0;  # this cluster number does not point to a cluster chain
      } elsif (0x0FFFFFFF == $tmp) {
         $clus = $nxt;  # this is a one-cluster chain
         @hoplist = ($clus);
         $hops2last = 0;
      } else {
         # this cluster number is the start of a 2-or-more-cluster chain
         do {
            $clus = $nxt;
            $hops2last++;
            push( @hoplist, $clus );
            $nxt = &Fat1Entry( $clus );
         } until (($nxt <= 0) || ($nxt >= 0x0FFFFFF7));
         if ((0xFFFFFFF8 & $nxt) != 0x0FFFFFF8) {
            # cluster chain did not end properly -- should end in -0x0FFFFFFF
            $cluster_chain_missing = 1 + $hops2last;  # side-effect error indicator result
         }
         if ($nxt != 0x0FFFFFFF) {
            printf "dbg-ERR: LastClust(0x0%X) bad-term\n", $frst;
            printf "  nxt=0x0%X,  nxt & 0x0FFFFFF8 = 0x0%X\n", $nxt, (0x0FFFFFF8 & $nxt);
         }
      }
      # if ($hops2last != $#hoplist) { print "ERR: LastClust bug: hops2last=$hops2last != " . $#hoplist . "\n"; }
   }
   return $clus;
} # sub LastClust


# open the drive or image file:
if ($o_reconcilewrite) {
   print "Opening '$drive' for reading and writing.\n";
   $rc = open( DISKIMAGE, "+<$drive" );
} else {
   print "Opening '$drive' for reading only.\n";
   $rc = open( DISKIMAGE, "<$drive" );
}
if (!$rc) {
   print "ERROR: Could not open '$drive', $!\n";
   &quit;
}
binmode DISKIMAGE;


# -h option searches for likely boot sectors
if ($o_search4bootsect) {
   if (!defined $lba_offset2) {
      print "ERR: -h w/ undefined 2nd sector number\n";  # should be impossible
   } else {
      # user specified a range of sector numbers
      for ($lba_ofs=$lba_offset1; $lba_ofs <= $lba_offset2; $lba_ofs++) {
         $score = &test4bootsect( $lba_ofs );
         if ($score > 6) {
            print "\nPossible boot sector found at LBA $lba_ofs:\n";
            &read_check_and_display_bootsect( $lba_ofs );
         }
      }
   }
   &quit(0);
}


# user specified just one sector number
&read_check_and_display_bootsect( $lba_offset1 );

if (512 != $b_bytesPerSector) {
   print "ERR: $b_bytesPerSector bytes per sector?!?  We assume 512.\n";
   &quit;
}

# debug
# for ($i=600; $i<512; $i++) {
#    $tmp = &FatEntry($i);
#    printf "%03Xh: %s -> %s", $i * 4, &DispFatEnt($i), &DispFatEnt($tmp);
#    if ($tmp > 0) {
#       $tmp = &LastClust($i);
#       printf " ...[%d hops]... -> %s", $hops2last, &DispFatEnt($tmp);
#       $tmp = &FatEntry($tmp);
#       printf " -> %s", &DispFatEnt($tmp);
#    }
#    print "\n";
# }


# Is the FAT32 filesystem "dirty"?
# The Dirty indicator is stored in the high 2 bits of the 2nd FAT entry:

$dirtyness = &UncheckedFatEntry(1);
if (($dirtyness & 0x08000000) != 0) {
   print "Volume is clean.\n";
} else {
   print "*** Volume is dirty ***\n";
}
if (($dirtyness & 0x04000000) != 0) {
   print "No hard errors were seen since last mount.\n";
} else {
   print "*** HARD ERRORS HAVE BEEN SEEN ON THIS VOLUME RECENTLY! ***\n";
}


# Our inode table is represented as a set of arrays (since Perl doesn't
# have good support for arrays of structures).  We have several arrays of
# things looked up by inode (file number).  Inode #0 is the root directory,
# so we initialize the table with just that single entry:
$in_Shortname[0] = '';  # short filename, from the directory entry
$in_Longname[0] = '';  # long filename, only included if '-n' specified
$in_ParentDir[0] = 0;  # inode of the directory it is in
$in_FirstClust[0] = $b_clusterOfRootDirStart;  # starting cluster number (usually from the directory entry, but for the root directory the starting cluster is from the boot sector)
$in_FType[0] = 'd';  # 'f' for plain file, 'd' for directory, 'v' for volume label
# in_FSize[] has file size in bytes, usually from the directory entry (except in this case):
&LastClust( $b_clusterOfRootDirStart );
if (!$o_minmem) {
   $in_FSize[0] = (1 + $hops2last) * $b_sectorsPerCluster * 512;  # $hops2last was set by &LastClust()
}

# When reading the directory info, we scan the table from the beginning,
# searching for additional subdirectories to scan.  When we find one, we
# call &Read1dir() to add its entries to the inode table.
$nxt_inode_to_examine = 0;


# Perl pack/unpack format representing the structure of a regular 32-byte directory entry:
$DirEntry_fmt =
  'a11' . # 000h 11 bytes  Short filename (8.3, but the '.' is implied, not stored)
  'C'   . # 00Bh  1 byte   Attributes: 1=R, 2=H, 4=S, 8=volumeID, 10h=D, 20h=A, 0Fh=longName
  'C'   . # 00Ch  1 byte   reserved
  'C'   . # 00Dh  1 byte   Create time 10ths, 0-199.
  'S'   . # 00Eh  2 bytes  Create time (units = 2 sec.)
  'S'   . # 010h  2 bytes  Create date (DOS fmt)
  'S'   . # 012h  2 bytes  Last access date (DOS fmt)
  'S'   . # 014h  2 bytes  High 16 bits of 32-bit first cluster no.
  'S'   . # 016h  2 bytes  Last write time (units = 2 sec.)
  'S'   . # 018h  2 bytes  Last write date (DOS fmt)
  'S'   . # 01Ah  2 bytes  Low 16 bits of 32-bit first cluster no.
  'L';    # 01Ch  4 bytes  File size, in bytes

# Perl pack/unpack format representing the structure of a long-name directory entry:
$DirLnameEntry_fmt =
  'C'   . # 000h  1 byte   Fragment number
  'S5'  . # 001h 10 bytes  Five unicode characters of the long file name
  'C'   . # 00Bh  1 byte   Attributes: must be 1Fh (=longName)
  'C'   . # 00Ch  1 byte   reserved (must be zero)
  'C'   . # 00Dh  1 byte   Checksum of corresponding short file name
  'S6'  . # 00Eh 12 bytes  Six more unicode characters of the long file name
  'S'   . # 01Ah  2 bytes  ClustLo must be zero (corresponds to low 16 bits of cluster no. in regular dir entry)
  'S2';   # 01Ch  4 bytes  Two more unicode characters of the long file name



# rotate a byte to the right by one bit
sub RotByteRight {
   local( $i ) = @_;
   local( $result ) = 0xFF & ($i >> 1);
   if ($i & 1) {
      $result |= 0x80;
   }
   return $result;
}


# Compute checksum of 11-character (8+3) short file name, as represented in
# the short directory entry.  The "." is implied but is not part of the
# input string and not part of the calculation, however the padding blanks
# ARE part of the calculation.  (This checksum is calculated from the short
# file name, and stored in each of the corresponding long file name entries,
# as a way of validating that the long file name entries really do go with
# that short file name.)
sub CkSum {
   local( $sfname ) = @_;
   local( $b, $i, @sfnam, $result );
   $result = 0;
   @sfnam = split( //, $sfname );
   # print "dbg0: split('$sfname') = " . ($#sfnam+1) . " parts\n";
   for ($i = 0; $i < 11; $i++) {
      $b = ord( $sfnam[$i] );
      $result = (0xFF & (&RotByteRight( $result ) + $b));
   }
#   printf "dbg: sfname=\"%s\", cksum=%02X\n", $sfname, $result;
   return $result;
}


# return a 32-byte directory entry from specified offset into first input parameter
sub selDirEnt {
  local( $dblk, $ofs ) = @_;
  if (!defined $dblk) {
     print "ERR: selDirEnt(undef,...)\n";
  }
  if (!defined $ofs) {
     $ofs = 0;
  }
  $dblk = substr( $dblk, $ofs, 32 );
  if (!defined $dblk) {
     print "ERR: dirent undefined [2].\n";
  }
  return $dblk;
}


sub unpackDirEnt {
  local( $direntry ) = @_;
  local( @results );
  if (!defined $direntry) {
     print "ERR: dirent undefined.\n";
     &quit(2,1);
  }
  @results = unpack( $DirEntry_fmt, $direntry );
  return @results;
}


# $x = "\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0A\x0B\x0C\x0D\x0E\x0F\x10\x11\x12\x13\x14\x15\x16\x17\x18\x19\x1A\x1B\x1C\x1D\x1E\x1F\x20";
# &unpackDirEnt_lname($x);
# exit 0;

# side-effect results are:  $ln_cksum, $ln_fragno, $ln_frag_looks_ok, and $ln_lname_piece
sub unpackDirEnt_lname {
  local( $direntry ) = @_;
  local( @tmp );
#  print "unpackDirEnt_lname:\n";
  if (!defined $direntry) {
     print "ERR: unpackDirEnt_lname(undef).\n";
  }
#  &dump_hex_and_ASCII( $direntry );
  @tmp = unpack( $DirLnameEntry_fmt, $direntry );
  if ($#tmp != 17) {
     die "ERR: unpackDirEnt_lname(), " . ($#tmp+1) . " parts instead of 18\n";
  }
#  print "dbg3: ";
#  for (local($i)=0; $i <= $#tmp; $i++) {
#     printf " %04X", $tmp[$i];
#  }
#  print "\n";
  local($ln_clustLo) = splice( @tmp, 15, 1 );
  $ln_cksum = splice( @tmp, 8, 1 );
  local($ln_resv) = splice( @tmp, 7, 1 );
  local($ln_attr) = splice( @tmp, 6, 1 );
  $ln_fragno = splice( @tmp, 0, 1 );
  $ln_frag_looks_ok = 1;
  if ((($d_attr & 0x3F) != 0xF) && ($d_attr != 0)) {
     printf "ERR: attr=0x%02X (should be 0x0F)\n", $ln_attr;
  }
  if ($ln_resv != 0) {
     if ($o_debugmode) {
        print "dbg-warn: ln_resv=$ln_resv (should be 0)\n";
     }
     $ln_frag_looks_ok = 0;
  }
  if ($ln_clustLo != 0) {
     if ($o_debugmode) {
        print "dbg-warn: ln_clustLo=$ln_clustLo (should be 0)\n";
     }
     $ln_frag_looks_ok = 0;
  }
  if ($#tmp != 12) {
     print "ERR: splice didn't remove?  " . $#tmp . " should be 12.\n";
  }
#  print "dbg4.1: ";
#  for (local($i)=0; $i <= $#tmp; $i++) {
#     printf " %04X", $tmp[$i];
#  }
#  print "\n";
  # Delete Unicode characters which do not map to legal ASCII (and also delete trailing 0-byte):
  for (local($i) = $#tmp; $i >= 0; $i--) {
     if (($tmp[$i] < 32) || ($tmp[$i] > 126)) {
        splice( @tmp, $i, 1 );
     }
  }
  $ln_lname_piece = pack( "c*", @tmp );
#  print "dbg4.2: ";
#  for (local($i)=0; $i <= $#tmp; $i++) {
#     printf " %04X", $tmp[$i];
#  }
#  print "\n";
#  &dump_hex_and_ASCII( $ln_lname_piece );

} # sub unpackDirEnt_lname


$dirent = $thirtytwo0s;  # $dirent is a side-effect result of &process1dirent

# Process a directory entry (but don't add it to the inode tables).
# Global side-effect results are:
#   $dirent = the 32-byte raw directory entry
#   $d_sfname = short filename
#   $d_sfname_raw = 11-byte short file name as it appears in the directory
#   $d_fsize = file size in bytes
#   $d_clust = first cluster number
#   $d_attr = attributes
#   $d_ftype = based on attributes: 'f', 'd', or other  ("\0" for null entry)
#   $d_deleted = 1 if entry represents a deleted file, 0 otherwise
sub process1dirent {
  local( $direntblk, $ofs ) = @_;
  $dirent = &selDirEnt( $direntblk, $ofs ); # sets global side-effect result
  local( $d_resv1, $d_ctimeLo, $d_ctimeHi, $d_cdate, $d_adate, $d_clustHi, $d_wtime, $d_wdate, $d_clustLo );
  ( $d_sfname_raw, $d_attr, $d_resv1, $d_ctimeLo, $d_ctimeHi, $d_cdate, $d_adate,
    $d_clustHi, $d_wtime, $d_wdate, $d_clustLo, $d_fsize ) = &unpackDirEnt( $dirent );
  # if (!defined $d_fsize) {
  #    $d_ftype = '?';
  #    return;
  # }
  $d_clust = $d_clustLo + ($d_clustHi << 16);
  $d_sfname = $d_sfname_raw;

  $d_deleted = 0;
  local( $firstch ) = substr($dirent, 0, 1);
  if ("\xE5" eq $firstch) {
     $d_deleted = 1;
  }
  if (($thirtytwo0s eq $dirent) || ($thirtyone0s eq $dirent)) {
     # null directory entry (unused space at end of last directory cluster)
     $d_ftype = "\0";
     undef $d_sfname;
     return;
  }

  if (($d_attr & 0x3F) == 0xF) {
     # it's part of a long file name
     $d_ftype = 'l';
     &unpackDirEnt_lname( $dirent );
  } elsif (   (0==$d_attr)
           && (0==$d_clustLo)
           && (($firstch eq "\xE5") || ($firstch eq "\0"))
           && (   ("\0" eq substr($dirent,2,1))
               || ("\0" eq substr($dirent,4,1))
               || ("\0" eq substr($dirent,6,1))
               || ("\0" eq substr($dirent,8,1))
               || ("\0" eq substr($dirent,10,1)) )
           && (   ("\0" ne substr($dirent,1,1))
               || ("\0" ne substr($dirent,3,1))
               || ("\0" ne substr($dirent,5,1))
               || ("\0" ne substr($dirent,7,1))
               || ("\0" ne substr($dirent,9,1)) ) ) {
     # it's part of a deleted long file name
     $d_deleted = 1;
     $d_ftype = 'l';
     &unpackDirEnt_lname( $dirent );
  } elsif (($d_attr & 8) == 8) {
     # volume label
     $d_ftype = 'v';
     $d_sfname = $d_sfname_raw;
     $d_sfname =~ s/[\s\0]+$//;
  } else {
     if (($d_attr & 0x10) == 0x10) {
        $d_ftype = 'd';  # subdirectory
     } elsif (0 == $d_clust) {
        $d_ftype = 'z';  # zero-length file
     } else {
        $d_ftype = 'f';  # plain file
     }
     local( $base, $ext ) = unpack( "a8a3", $d_sfname_raw );
     if ($firstch eq "\xE5") {
        # deleted entry
        $d_deleted = 1;
        $base = '?' . substr($base, 1);
     } elsif ($firstch eq "\0") {
        $base = '?';
     }
     $base =~ s/[\s\0]+$//;
     if ($base =~ /^\x05/) {
        # special hack for Japanese Kanji language
        $base = "\xE5" . substr($base,1);
     }
     $ext =~ s/[\s\0]+$//;
     # print "dbg: base='$base', ext='$ext'\n";
     if ('' ne $ext) {
        $d_sfname = $base . '.' . $ext;
     } else {
        $d_sfname = $base;
     }
  }
}#process1dirent


# return the index of the first blank in the input parameter string (or the length of the string if it contains no blanks)
sub firstblank {
   local( $istr ) = @_;
   local( $i, @chrs );
   local( $result );
   @chrs = split(//,$istr);
   $result = 1 + $#chrs;
   for ($i = 0; $i <= $#chrs; $i++) {
      if (' ' eq $chrs[$i]) {
         $result = $i;
         last;
      }
   }
   return $result;
}

# printf "dbg: firstblank( ' 123 45678' = %d\n", &firstblank(' 123 45678');
# printf "dbg: firstblank( '123 4 5678' = %d\n", &firstblank('123 4 5678');
# printf "dbg: firstblank( '123 45678 ' = %d\n", &firstblank('123 45678 ');
# printf "dbg: firstblank( '12345678' = %d\n", &firstblank('12345678');
# printf "dbg: firstblank( '' = %d\n", &firstblank('');
# printf "dbg: firstblank( '1' = %d\n", &firstblank('1');
# printf "dbg: firstblank( ' ' = %d\n", &firstblank(' ');
# printf "dbg: firstblank( '  ' = %d\n", &firstblank('  ');


# return the index of the last non-blank in the input parameter string (or -1 if it contains no non-blanks)
sub lastnonblank {
   local( $istr ) = @_;
   local( $i, @chrs );
   local( $result );
   @chrs = split(//,$istr);
   $result = -1;
   for ($i = $#chrs; $i >= 0; $i--) {
      if (' ' ne $chrs[$i]) {
         $result = $i;
         last;
      }
   }
   return $result;
}

# printf "dbg: lastnonblank( ' 123 45678' = %d\n", &lastnonblank(' 123 45678');
# printf "dbg: lastnonblank( '123 4 5678' = %d\n", &lastnonblank('123 4 5678');
# printf "dbg: lastnonblank( '123 45678 ' = %d\n", &lastnonblank('123 45678 ');
# printf "dbg: lastnonblank( '12345678' = %d\n", &lastnonblank('12345678');
# printf "dbg: lastnonblank( '' = %d\n", &lastnonblank('');
# printf "dbg: lastnonblank( '1' = %d\n", &lastnonblank('1');
# printf "dbg: lastnonblank( ' ' = %d\n", &lastnonblank(' ');
# printf "dbg: lastnonblank( '  ' = %d\n", &lastnonblank('  ');



# Given a 32-byte object, determine whether or not it looks like a directory entry.
# Rating is high if it looks like a directory entry, negative if it can't be one,
# zero if it is all zero-bytes (which is common at the end of the last sector of
# a directory).
sub rate_1dirent {
  local( $dirent ) = @_;
  local( $result ) = 0xDABDAB;
  local ( $d_sfname_raw, $d_attr, $d_resv1, $d_ctimeLo, $d_ctimeHi, $d_cdate, $d_adate,
          $d_clustHi, $d_wtime, $d_wdate, $d_clustLo, $d_fsize );
  local($firstch, $firstch2, @namechrs, $blankseen);

  if (32 != length($dirent)) {
     print "ERR: rate_1dirent\n";
     &quit(2,1);
  }

  if ($o_debugmode > 1) {
     print "dbg: Enter rate_1dirent for:\n";
     &dump_hex_and_ASCII( $dirent );
  }

  $d_deleted = 0;
  $firstch = substr($dirent, 0, 1);
  if ("\xE5" eq $firstch) {
     $d_deleted = 1;
  }
  if (($thirtytwo0s eq $dirent) || $d_deleted) {
     # null directory entry (unused space at end of last directory cluster)
     $d_ftype = "\0";
     $result = 0;
  }

  # printf "dbg[2], result=0x0%X\n", $result;

  if (0xDABDAB == $result) {

     ( $d_sfname_raw, $d_attr, $d_resv1, $d_ctimeLo, $d_ctimeHi, $d_cdate, $d_adate,
       $d_clustHi, $d_wtime, $d_wdate, $d_clustLo, $d_fsize ) = &unpackDirEnt( $dirent );

     if ($d_attr == 0xF) {
        # it's part of a long file name
        $d_ftype = 'l';
        &unpackDirEnt_lname( $dirent );  # side-effect results are:  $ln_cksum, $ln_fragno, $ln_frag_looks_ok, and $ln_lname_piece
        if ($ln_frag_looks_ok) {
           $result = 3;
        } else {
           $result = -50;
        }
     }
  }

  # printf "dbg[3], result=0x0%X\n", $result;

  if (0xDABDAB == $result) {
     if (!defined $d_clustHi) {
        print "ERR: how tha heck can d_clustHi be undefined here??\n";
        &quit(3,1);
     }
     $d_clust = $d_clustLo + ($d_clustHi << 16);
     $d_sfname = $d_sfname_raw;
     if (   (0==$d_attr)
           && (0==$d_clustLo)
           && (($firstch eq "\xE5") || ($firstch eq "\0"))
           && (   ("\0" eq substr($dirent,2,1))
               || ("\0" eq substr($dirent,4,1))
               || ("\0" eq substr($dirent,6,1))
               || ("\0" eq substr($dirent,8,1))
               || ("\0" eq substr($dirent,10,1)) )
           && (   ("\0" ne substr($dirent,1,1))
               || ("\0" ne substr($dirent,3,1))
               || ("\0" ne substr($dirent,5,1))
               || ("\0" ne substr($dirent,7,1))
               || ("\0" ne substr($dirent,9,1)) ) ) {
        # it's part of a deleted long file name
        $d_deleted = 1;
        $d_ftype = 'l';
        &unpackDirEnt_lname( $dirent );
        $result = 0;
     } elsif (($d_attr == 8) || ($d_attr == 0x28)) {
        # volume label... not legal except in root directory
        $d_ftype = 'v';
        $result = -10;
     } elsif (($d_attr & 0xC8) != 0) {
        # illegal attribute
        $result = -51;
     }
  }

  # printf "dbg[4], result=0x0%X\n", $result;

  if (0xDABDAB == $result) {
     if (($d_attr & 0x10) == 0x10) {
        $d_ftype = 'd';  # subdirectory
     } else {
        $d_ftype = 'f';  # plain file
     }
     # printf "dbg[4.1], result=0x0%X, d_ftype=$d_ftype\n", $result;
     if (($firstch eq "\xE5") || ($firstch eq "\0")) {
        # deleted entry
        $d_deleted = 1;
        if ($firstch eq "\xE5") {
           $result = 1;
        } else {
           $result = 0;
        }
     } elsif (   (0 == $d_clust)
              && (($firstch eq ".") || ((0 == $d_fsize) && ('f' eq $d_ftype))) ) {
        $result = 2;
        # the only three circumstances in which the cluster number may be zero are:
        # 1) A zero-length file may have a zero cluster number, and
        # 2) The ".." directory entry apparently may have a zero cluster number (oddly enough), and
        # 3) Deleted entries may have zero cluster numbers (but that's handled above)
     } elsif (!&is_valid_clusterno($d_clust)) {
        $result = -52;
        # printf "dbg[4.2], rate_1dirent, clust = 0x0%X is invalid, result=%d\n", $d_clust, $result;
        # &dump_hex_and_ASCII( $dirent );
        @namechrs = split( //, $d_sfname_raw );
        # printf "          firstch=%d='%s', d_ftype=$d_ftype, namechrs='%s'\n", ord($namechrs[0]), $namechrs[0], join('/',@namechrs);
        # printf "          d_clustH=0x0%X, d_clustLo=0x0%X, d_fsize=%s, d_wtime=0x0%X, d_wdate=0x0%X\n", $d_clustHi, $d_clustLo, &commas5($d_fsize), $d_wtime, $d_wdate;
     } else {
        # printf "dbg[4.3], result=0x0%X\n", $result;
        # ($firstch2, @namechrs) = split( //, $d_sfname_raw );
        ($firstch2, @namechrs) = unpack( "aaaaaaaaaaa", $d_sfname_raw );
        # printf "dbg[4.4], result=0x0%X, firstch=%d=%d, d_ftype=$d_ftype, namechrs='%s'\n", $result, ord($firstch), ord($firstch2), join('/',@namechrs);
        # other than the first character, none of the remaining characters may be "\0":
        $blankseen = 0;
        foreach $ch (@namechrs) {
           if ("\0" eq $ch) {
              $result = -53;
              last;
           } elsif (' ' eq $ch) {
              $blankseen = 1;
           }
        }
        if (0xDABDAB == $result) {
           # blanks can only pad the end of the name or extension
           $result = 4;
           if ($blankseen) {
              $base7 = substr( $d_sfname_raw, 1, 7 );
              $ext = substr( $d_sfname_raw, 8, 3 );
              if (&firstblank($base7) < &lastnonblank($base7)) {
                 $result = -30;
              } elsif (&firstblank($ext) < &lastnonblank($ext)) {
                 $result = -31;
              } else {
                 $result = 5;
              }
           } else {
              $result = 6;
           }
        }
     }
  }

  if ($o_debugmode || ($result < 0)) {
     # printf "dbg[5], result=0x0%X\n", $result;
     print "dbg: rate_1dirent returns result = $result for:\n";
     &dump_hex_and_ASCII( $dirent );
     # print "--- Exit rate_1dirent ---\n";
  }
  return $result;
} # sub rate_1dirent


# Rate an entire cluster or sector as to the likelihood that it is part of a directory:
# Positive means it is part of a directory, negative means it isn't.
sub rate_1dircluster {
   local( $clust, $clustno ) = @_;
   local( $result ) = 0;
   local( $direntry );
   local( $ofs );
   for ($ofs=0; $ofs < length($clust); $ofs+=32) {
      $direntry = substr( $clust, $ofs, 32 );
      $result += &rate_1dirent( $direntry );
   }

   printf "dbg: rate_1dircluster returns result=$result";
   if (defined $clustno) {
      printf " for cluster no. %s = 0x0%X (starting at LBA %s)",
          &commas5($clustno), $clustno, &commas5(&cluster_to_lba($clustno));
   }
   print "\n";

   return $result;
} # sub rate_1dircluster


# add a directory entry to the inode tables
sub add1dirent {
  local( $parent_inode, $d_sfname, $d_fsize, $d_clust, $d_ftype, $lname ) = @_;
  local( $inode ) = $#in_Shortname + 1;
  $in_Shortname[$inode] = $d_sfname;  # short name
  $in_ParentDir[$inode] = $parent_inode;  # inode of the directory it is in
  $in_FirstClust[$inode] = $d_clust;  # starting cluster number
  $in_FType[$inode] = $d_ftype;  # 'f' for plain file, 'd' for directory
  if ($o_longnames) {
     $in_Longname[$inode] = $lname;
  }
  if ($o_debugmode) {
     printf "dbg: add1dirent[$inode]( $parent_inode, '%s', $d_fsize, 0x0%X, $d_ftype, '%s' )\n",
            $d_sfname, $d_clust, $lname;
  }
  if (!$o_minmem) {
     $in_FSize[$inode] = $d_fsize;
  }
}


sub inode2path {
  local( $inode ) = @_;
  local( $parent_inode ) = $in_ParentDir[$inode];
  local( $i, $k, $result);
  $result = '';
  $k = 0;  # protection against infinite loop
  if ($inode > $#in_Shortname) {
     $result = "***ERROR***";
  } else {
     if ($inode > 0) {
        $result = $in_Shortname[$inode];
        while ($k++ < 100) {
           $inode = $in_ParentDir[$inode];
           if (0 == $inode) {
              last;
           }
           $result = $in_Shortname[$inode] . "\\" . $result;
        }#while
     }
     $result = "\\" . $result;
  }
  return $result;
}


# When stepping through a directory, we "accumulate" the pieces of a long file
# name, until we encounter something other than the next piece of the file name
# or the corresponding short file name.  If that happens, we discard whatever
# long file name information we've thus far accumulated, by calling this
# subroutine.
sub reset_accumulating_lname {
  $accumulating_lname = '';
  $accumulating_cntr = 0;
  $prev_ftype = ' ';
  $prev_deleted = -1;
  $prev_cksum = -1;
  $prev_fragno = -1;
}


# This subroutine processes a directory (folder).
sub Read1dir {
  local( $clust1, $parent_inode ) = @_;
  local( $lastcluster ) = &LastClust( $clust1 );
  local( $DirSizeInClusters ) = 1 + $hops2last;  # $hops2last was set by &LastClust()
  # local( $DirSizeInSectors ) = $b_sectorsPerCluster * $DirSizeInClusters;
  local( $clustno ) = $clust1;
  local( $lba1, $cksum, $printthis1, $deleted_mark, $i );
  local($cluster);
  local( $list_lnames ) = (($o_rawlistfiles >= 2) || ($o_listdeletedfiles >= 3) || $o_verbose || $o_debugmode);

  if (! &is_valid_clusterno($clustno)) {
     printf "ERROR: Read1dir( cluster=%s=0x%08X, inode=$inode ); bad cluster no.!\n", &commas5($clustno), $clustno;
     return;
  }

  if ($o_debugmode || $o_verbose) {
     print "clust1 = $clust1 -> " . &DispFatEnt( &Fat1Entry($clust1) );
     print ", lastcluster = $lastcluster,  size = $DirSizeInClusters clusters ($b_sectorsPerCluster sectors/cluster)\n";
  }

  # find the full path of the parent directory
  local( $parentdir ) = &inode2path( $parent_inode );
  if (length($parentdir) > 1024) {
     print "ERROR: path too long, possible recursive directory structure, skipping directory:\n " . $parentdir . "\n";
     return;
  }

  &reset_accumulating_lname;

  while ((0x0FFFFFFF != $clustno) && &is_valid_clusterno($clustno)) {
     $lba1 = &cluster_to_lba($clustno);
     if ($o_debugmode || $o_verbose) {
        printf "Cluster %d=0x%08X, which is LBA $lba1=0x0%X, which is byte offset %s=0x0%X\n",
               $clustno, $clustno, $lba1, &commas(512 * $lba1), (512 * $lba1);
     }
     $cluster = &bigread1( $lba1 * 512, $b_sectorsPerCluster );
     if ($o_debugmode) {
        printf "length(cluster) = %d.\n", length($cluster);
        # &dump_by_sector( $cluster );
        if ($o_debugmode >= 2) {
           &dump_hex_and_ASCII( $cluster );
        }
     }

     if ($o_debugmode || $o_rawlistfiles) {
        &rate_1dircluster( $cluster, $clustno );  # dbg
     }

     for ($i=0; $i < length($cluster); $i+=32) {
        &process1dirent( $cluster, $i, $parent_inode );  # $dirent and $d_* are side-effect results
        $deleted_mark = sprintf("%5d ",$i);
        if ($d_deleted) {
           $deleted_mark .= '*D*';
        } else {
           $deleted_mark .= '   ';
        }
        if (($thirtytwo0s eq $dirent) || ($thirtyone0s eq $dirent)) {
           if ($o_rawlistfiles > 1) {
              printf "%s           <ZEROS>\n", $deleted_mark;
           }
        } elsif (defined $d_sfname) {
           # found a directory entry
           if ($o_rawlistfiles || $o_longnames || $o_listdeletedfiles || $o_verbose || $o_debugmode) {
              # piece together the long name fragments
              if ('l' eq $d_ftype) {
#                 $tmp = &selDirEnt( $cluster, $i );
#                 &dump_hex_and_ASCII( $tmp );
                 if ('' ne $accumulating_lname) {
                    if (    ($ln_cksum != $prev_cksum)
                         || ($prev_ftype ne $d_ftype)
                         || ($d_deleted != $prev_deleted)
                         || !$ln_frag_looks_ok) {
                        # the checksums & deleted-ness must match for all the pieces of a long name, whether deleted or not
                        # print "dbg1: ln_cksum=$ln_cksum, prev_cksum=$prev_cksum, d_ftype=$d_ftype, prev_ftype=$prev_ftype, d_deleted=$d_deleted, prev_deleted=$prev_deleted, ln_frag_looks_ok=$ln_frag_looks_ok\n";
                        &reset_accumulating_lname;
                    } elsif (!$d_deleted) {
                        if (($ln_fragno & 0x40) || ($ln_fragno != (($prev_fragno & 0x3F) - 1))) {
                           # for non-deleted long names, the first fragment number must be OR'd with 0x40, and they count down to 1
                           # printf "dbg2: ln_fragno=0x%02X, prev_fragno=0x%02X\n", $ln_fragno, $prev_fragno;
                           &reset_accumulating_lname;
                        }
                    }
                 }
                 if ($ln_frag_looks_ok) {
                    if ('' ne $accumulating_lname) {
                       $accumulating_lname = $ln_lname_piece . $accumulating_lname;
                    } elsif ((0 != ($ln_fragno & 0x40)) || $d_deleted) {
                       $accumulating_lname = $ln_lname_piece;
                    } else {
                       # printf "dbg3: ln_fragno=0x%02X, d_deleted=%d\n", $ln_fragno, $d_deleted;
                       &reset_accumulating_lname;
                    }
                 }
                 if ('' ne $accumulating_lname) {
                    $prev_cksum = $ln_cksum;
                    $prev_fragno = $ln_fragno;
                    $prev_deleted = $d_deleted;
                    $prev_ftype = $d_ftype;
                    $accumulating_cntr++;
                 }
                 # if '-L' or '-d' was specified, then list the directory entry for the longname fragment
                 if (    ($list_lnames && !$d_deleted)
                      || ($o_rawlistfiles > 1)
                      || (    $d_deleted
                           && (    ($list_lnames && $o_listdeletedfiles)
                                || ($o_listdeletedfiles >= 2) ) ) ) {
                    printf "%s           <LNAME> (%02X)  %02X  %d  \"%s\"",
                           $deleted_mark, $ln_cksum, $ln_fragno, $ln_frag_looks_ok, $ln_lname_piece;
                    printf "%*s%s\n", 15-length($ln_lname_piece), '', $parentdir;
                    # printf "dbg: d_deleted=$d_deleted, deleted_mark='$deleted_mark'\n";
                 }
              } else {
                 # $d_ftype is not 'l'
                 $cksum = &CkSum( $d_sfname_raw );
                 if (($prev_deleted != $d_deleted) || (($cksum != $prev_cksum) && !$d_deleted)) {
                    # print "dbg8: prev_deleted=$prev_deleted, d_deleted=$d_deleted, cksum=$cksum, prev_cksum=$prev_cksum\n";
                    &reset_accumulating_lname;
                 }
                 # if '-L' or '-d' was specified, then list the directory entry
                 $printthis1 = 1;
                 if (   ($o_rawlistfiles && !$d_deleted)
                     || ($o_listdeletedfiles && $d_deleted)
                     || ($o_rawlistfiles > 1) ) {
                    if ('d' eq $d_ftype) {
                       print "$deleted_mark            <DIR> ";
                       # print " [dbg: siz=$d_fsize]";  # maybe always zero?
                    } elsif ('f' eq $d_ftype) {
                       printf "%s%11d   %02X  ", $deleted_mark, $d_fsize, $d_attr;
                    } elsif ('v' eq $d_ftype) {
                       printf "%s  Vol_Label:  %02X  ", $deleted_mark, $d_attr;
                    } elsif ('z' eq $d_ftype) {
                       printf "%s  EMPTY  %2d   %02X  ", $deleted_mark, $d_fsize, $d_attr;
                    } elsif ($o_rawlistfiles > 1) {
                       printf "%s  ftype='%s'?  %02X  ", $deleted_mark, $d_ftype, $d_attr;
                    } else {
                       $printthis1 = 0;
                    }
                    if ($printthis1) {
                       if ((!$list_lnames) && ($o_listdeletedfiles < 2)) {
                          printf "%07X  %12s   %s", $d_clust, $d_sfname, $parentdir;
                       } elsif ($d_deleted) {
                          printf "%07X  %12s          %s", $d_clust, $d_sfname, $parentdir;
                       } else {
                          printf "%07X  %12s (%02X)     %s", $d_clust, $d_sfname, $cksum, $parentdir;
                       }
                       print "\n";
                       if ('' ne $accumulating_lname) {
                          printf "      %s                           \"%s\"\n",
                                 substr($deleted_mark,6), $accumulating_lname;
                       }
                    }
                 }
              }
           }
           if (('d' eq $d_ftype) || ('f' eq $d_ftype)) {
              # files and subdirectories get added to the inode list, except for '.' and '..'
              if ((! $d_deleted) && ('.' ne $d_sfname) && ('..' ne $d_sfname)) {
                 &add1dirent( $parent_inode, $d_sfname, $d_fsize, $d_clust, $d_ftype, $accumulating_lname );
              }
           }
        }#if
        if ($o_rawlistfiles >= 3) {
           if ($thirtytwo0s ne $dirent) {
              if ($o_rawlistfiles >= 4) {
                  print " _0 _1 _2 _3 _4 _5 _6 _7 _8 _9 _A _B _C _D _E _F 10 11 12 13 14 15 16 17 18 19 1A 1B 1C 1D 1E 1F\n";
               }
               if ('l' ne $d_ftype) {
                  print " **______Base name______ ___Ext__ At 00 __ctime_ cdate adate ClusH wtime wdate ClusL ___Size____\n" .
                       " -----------------------.--------|--|--|-------- -----|-----|-----|----- -----|-----|-----------\n";
              } else {
                 print " F# _____Five_unicode_chars_____  1F 00 Ck ________Six_unicode_chars__________ zero_ _two_chars_\n" .
                       " --|-----,-----,-----,-----,-----|-- --|--|-----,-----,-----,-----,-----,-----|-----|-----,-----\n";
              }
           }
           foreach $ch (split(//,$dirent)) {
              $chI = 0x00FF & unpack("c",$ch);
              printf " %02X", $chI;
           }
           print "\n\n";
           # if ($o_rawlistfiles >= 4) {
           #    print "\n";
           # }
        }
        if ('l' ne $d_ftype) {
           # print "dbg7: d_ftype=$d_ftype != 'l'\n";
           &reset_accumulating_lname;
        }
     }#for
     $clustno = &Fat1Entry($clustno);  # next cluster in chain that makes up this directory
  }#while
}# sub Read1dir


if (!($o_lookup_sector | $o_listfiles | $o_rawlistfiles | $o_reconcilewrite | $o_correctFATfromDirs)) {
   &quit(0);
}


# From this point on, we don't need $fat2, so free the memory that it occupies:
undef $fat2;

print "Traversing folders on FAT32 partition... (this might take awhile)\n";

# if ($o_debugmode) {
   $timer{'read_all_dirs'} = 0.0;
   $timer{'read_all_dirs'} -= (times)[0];
# }


# Read the inode list, processing each entry.  For subdirectory
# entries (type 'd') we call Read1dir to read the subdirectory,
# which adds to the end of the inode list.  So by the time we've
# gotten to the end of the inode list, we've actually read all
# the directories in the volume.
for ($inode=$nxt_inode_to_examine; $inode <= $#in_Shortname; $inode++) {
   if ('d' eq $in_FType[$inode]) {
      if ($o_debugmode) {
         print "dbg: read directory '$in_Shortname[$inode]' (inode $inode)\n";
      }
      if (&is_valid_clusterno($in_FirstClust[$inode])) {
         &Read1dir( $in_FirstClust[$inode], $inode );
      } else {
         printf "Skipping folder '%s' (inode $inode) due to invalid start cluster, 0x0%X\n",
                $in_Shortname[$inode], $in_FirstClust[$inode];
      }
   }
   # if ($inode > 10000) {
   #    print "dbg: aborting at 10,000th inode.\n";
   #    last;
   # }
}

# if ($o_debugmode) {
   $timer{'read_all_dirs'} += (times)[0];
   printf "dbg: reading directories took %.2f seconds.\n", $timer{'read_all_dirs'};
   print "dbg: begin building cluster-to-inode hash lookup table...\n";
   $timer{'build_LastClust2inode'} = 0.0;
   $timer{'build_LastClust2inode'} -= (times)[0];
# }



if ($o_listfiles >= 3) {
   # '-l3'
   print
"             size  ---cluster(s)--- clust-term path\n" .
"        ---------  ---------------- ---------- ---------\n";
} elsif ($o_listfiles) {
   # '-l' or '-ll'
   print
"             size  ---cluster(s)--- path\n" .
"        ---------  ---------------- ---------\n";
}


$sum_file_sizes = 0.0;
$sum_clusters_used = 0;
$num_inodes_without_valid_clusters = 0;

# build hash for looking up inode from the last cluster number in a cluster chain
sub build_LastClust2inode {
   local( $inode, $clust );
   undef %LastClust2inode;
   for ($inode=0; $inode <= $#in_Shortname; $inode++) {
      $clust = $in_FirstClust[$inode];
      if (&is_valid_clusterno($clust)) {
         $LastClust2inode{ &LastClust($clust) } = $inode;
         $sum_clusters_used += (1 + $hops2last);
         if (!$o_minmem) {
            $sum_file_sizes += $in_FSize[$inode];
         }
      } else {
         # there's no valid last cluster number for this inode
         $num_inodes_without_valid_clusters++;
         # needed for $num_lost_chains calculation
      }
   }
} # sub build_LastClust2inode


&build_LastClust2inode;


# if ($o_debugmode) {
   $timer{'build_LastClust2inode'} += (times)[0];
   printf "dbg: building hash lookup table took %.2f seconds.\n",
           $timer{'build_LastClust2inode'};
# }


printf "%s files \& folders use %s clusters = %s KB on disk\n",
       &commas5(1 + $#in_Shortname),
       &commas5($sum_clusters_used),
       &commas5( $sum_clusters_used * ($b_sectorsPerCluster / 2));

if (!$o_minmem) {
   printf "Summed file & folder sizes for %s files = %.3F MB\n",
          &commas5(1 + $#in_Shortname), ($sum_file_sizes + 511) / (1024 * 1024);
}


# Next, built a lookup table which tells the starting cluster
# number for any cluster chain, from the final cluster number

# build last-to-first cluster lookup table
undef %last_to_first_cluster;  # table is side-effect result
sub build_Last2FirstTable {
   local(%hops_in_chain);
   local($seventyfifthdone) = int($highest_cluster_num / 74);  # each progress indicator dot represents about 1.33% done
   local($thousandthdone) = int($highest_cluster_num / 999);
   local( $frst, $nxt, $clus, $tmp );  # locals used by inlined &LastClust
   if ($thousandthdone < 3) {
      # should be impossible, since FAT32 file systems are supposed to have at least 65,526 clusters
      $thousandthdone = 3;
   }
   local( $clust );
   $timer{'build_onethousandth'} = 0.0;
   $timer{'build_onethousandth'} -= (times)[0];
   print ".";  # progress indicator
   for ($clust=2; $clust <= $highest_cluster_num; $clust++) {
      if ($clust == $thousandthdone) {
         $timer{'build_onethousandth'} += (times)[0];
         printf "\nEstimated time to build last-to-first lookup table = %.2f minutes.\n",
                ((999.0 / 60.0) * $timer{'build_onethousandth'});
         print ".";
      }
      if (($clust % $seventyfifthdone) == 0) {
         print ".";  # progress indicator
      }

      # $last = &LastClust( $clust );
      # **** BEGIN inlined &LastClust() ****
      $frst = $nxt = $clust;
      # side-effect results:
      $hops2last = -1;
      # $cluster_chain_missing = 0;
      # @hoplist = ();
      $tmp = &UncheckedFatEntry( $nxt );
      if ((0 == $tmp) || (0x0FFFFFF7 == $tmp)) {
         $clus = 0;  # this cluster number does not point to a cluster chain
      } elsif (0x0FFFFFFF == $tmp) {
         $clus = $frst;  # this is a one-cluster chain
         # @hoplist = ($clus);
         $hops2last = 0;
      } else {
         # this cluster number is the start of a 2-or-more-cluster chain
         do {
            $clus = $nxt;
            $hops2last++;
            # push( @hoplist, $clus );

            # optimization (skip over contiguous cluster chains), which cut runtime
            # for &build_Last2FirstTable on Gregory's HDD from over 3 hrs to just 60 sec:
            if ($clus == (1+$clust)) {
               $clust = $clus;
            }

            $nxt = &Fat1Entry( $clus );
         } until (($nxt <= 0) || ($nxt >= 0x0FFFFFF7));
         if ((0xFFFFFFF8 & $nxt) != 0x0FFFFFF8) {
            # cluster chain did not end properly -- should end in -0x0FFFFFFF
            # $cluster_chain_missing = 1 + $hops2last;  # side-effect error indicator result
         }
         if ($nxt != 0x0FFFFFFF) {
            printf "dbg-ERR/2: LastClust-inline(0x0%X) bad-term\n", $frst;
            printf "  nxt=0x0%X,  nxt & 0x0FFFFFF8 = 0x0%X\n", $nxt, (0x0FFFFFF8 & $nxt);
         }
      }
      $last = $clus;
      # **** END inlined &LastClust() ****

      if ($last) {
         if (defined $hops_in_chain{$last}) {
            if ($hops2last > $hops_in_chain{$last}) {
               $hops_in_chain{$last} = $hops2last;
               $last_to_first_cluster{$last} = $frst;
            }
         } else {
            $hops_in_chain{$last} = $hops2last;
            $last_to_first_cluster{$last} = $frst
         }
      }
   }#for
   print "\n";  # end of progress indicator
   print "exiting build_Last2FirstTable\n";
   undef %hops_in_chain;
}

$timer{'build_Last2FirstTable'} = 0.0;
$timer{'build_Last2FirstTable'} -= (times)[0];
print "Begin building last-to-first cluster lookup table...\n";
&build_Last2FirstTable;
$timer{'build_Last2FirstTable'} += (times)[0];
printf "dbg: done building last-to-first lookup table (it took %.2f seconds).\n",
       $timer{'build_Last2FirstTable'};


# The list of starting cluster numbers for all the cluster
# chains is simply the keys from the %last_to_first_cluster
# lookup table:

@list_of_last_clusters = (sort keys %last_to_first_cluster);
$num_chains = 1 + $#list_of_last_clusters;
# printf "num_chains = %s\n", &commas5($num_chains);



# We need to find the "top" lost folder(s).  The strategy is to:
# 1) Examine every lost folder, and build from it a list of the
# other recovered folders listed within it.  Then,
# 2) Eliminate each referenced folder from the list.  What's left
# are the top folders.


# First, calculate how many lost cluster chains there are, by
# comparing the total number of cluster chains to the number
# of cluster chains pointed to by all inode entries (directory
# entries)

$num_lost_chains = 0;
$num_chains_referenced_in_folders = (1 + $#in_Shortname) - $num_inodes_without_valid_clusters;
if ($num_chains == $num_chains_referenced_in_folders) {
   printf "There are no lost cluster chains.  (%s chains found)\n", &commas5($num_chains);
} elsif ($num_chains <= $num_chains_referenced_in_folders) {
   printf "ERROR: fewer cluster chains than inodes!  Must be cross-linked files.  (%s chains, %s inodes)\n",
          &commas5($num_chains), &commas5(1+$#in_Shortname);
} else {
   $num_lost_chains = $num_chains - $num_chains_referenced_in_folders;
   printf "Warning: %s lost cluster chains detected.  (%s chains, but only %s inodes w/ cluster chains)\n",
          &commas5($num_lost_chains), &commas5($num_chains), &commas5($num_chains_referenced_in_folders);
}


# Next, build a list of lost cluster chains -- the cluster chains
# that aren't pointed to by any entries in %LastClust2inode

@lost_chains = ();

# if ($num_lost_chains) {
   foreach $last (@list_of_last_clusters) {
      if (!defined $LastClust2inode{$last}) {
         $clust = $last_to_first_cluster{$last};
         push( @lost_chains, $clust );
      }
   }
   if ($num_lost_chains != (1 + $#lost_chains)) {
      printf "ERR: num_lost_chains=%s != %s\n", &commas5($num_lost_chains), &commas5(1+$#lost_chains);
   }
# }


@lost_chains = sort @lost_chains;  # sort in order of starting cluster no.

if ($o_debugmode) {
   $tmp = sprintf("%07X", $lost_chains[0]);
   for ($clust=1; $clust<=$#lost_chains; $clust++) {
      $tmp .= sprintf(", %07X",$lost_chains[$clust]);
   }
   print "Starting clusters for lost cluster chains are:\n $tmp\n";
}


# Next, determine which of the lost cluster chains are lost folders,
# and make a list:

@lost_folders = ();
foreach $clust (@lost_chains) {
   $cluster = &fetchCluster( $clust );
   if (   (".          " eq substr($cluster,0,11))
       && ("..         " eq substr($cluster,32,11))
       && (&rate_1dircluster($cluster,$clust) > 0)  ) {
      push( @lost_folders, $clust );
      printf "dbg: Found a lost folder at cluster 0x0%X = %s\n", $clust, &commas5($clust);
   }
}
printf "Found %s lost folders.\n", &commas5( 1 + $#lost_folders );


# Next, create the %lost_dirs lookup table, for looking up a starting cluster to
# quickly determine whether or not it is actually the starting cluster of a lost
# directory.
undef $lost_dirs;
foreach $clust (@lost_folders) {
   $lost_dirs{$clust} = '';
}


undef %lost_file_counts;  # table we'll build, listing the number of files & folders contained within each lost dir
$file_count = 0;  # side-effect result

# This subroutine reads a "lost" folder, searching for lost sub-folders.
# Result is a list of subfolder starting cluster numbers.  Only other lost
# folders will be included in the list, and only if they are not marked
# as deleted.
# Also sets global $file_count as a side-effect result.
sub Read1lostdir {
  local( $clust1 ) = @_;
  local( @result ) = ();
  $file_count = 0;
  local( $lastcluster ) = &LastClust( $clust1 );
  local( $DirSizeInClusters ) = 1 + $hops2last;  # $hops2last was set by &LastClust()
  # local( $DirSizeInSectors ) = $b_sectorsPerCluster * $DirSizeInClusters;
  local( $clustno ) = $clust1;
  local( $lba1, $deleted_mark, $i );
  local($cluster);

  if (! &is_valid_clusterno($clustno)) {
     printf "ERR: Read1lostdir( cluster=%s=0x%08X, inode=$inode ); bad cluster no.!\n", &commas5($clustno), $clustno;
     return @result;
  }

  if ($o_debugmode || $o_verbose) {
     print "Read1lostdir: clust1 = $clust1 -> " . &DispFatEnt( &Fat1Entry($clust1) );
     print ", lastcluster = $lastcluster,  size = $DirSizeInClusters clusters ($b_sectorsPerCluster sectors/cluster)\n";
  }

  while ((0x0FFFFFFF != $clustno) && &is_valid_clusterno($clustno)) {
     $lba1 = &cluster_to_lba($clustno);
     if ($o_debugmode || $o_verbose) {
        printf "Read1lostdir: Cluster %d=0x%08X, which is LBA $lba1=0x0%X, which is byte offset %s=0x0%X\n",
               $clustno, $clustno, $lba1, &commas(512 * $lba1), (512 * $lba1);
     }
     $cluster = &bigread1( $lba1 * 512, $b_sectorsPerCluster );
     if ($o_debugmode) {
        printf "length(cluster) = %d.\n", length($cluster);
        # &dump_by_sector( $cluster );
        if ($o_debugmode >= 2) {
           &dump_hex_and_ASCII( $cluster );
        }
     }
     if ($o_debugmode) {
        &rate_1dircluster( $cluster, $clustno );  # dbg
     }

     for ($i=0; $i < length($cluster); $i+=32) {
        &process1dirent( $cluster, $i, $parent_inode );  # $dirent and $d_* are side-effect results
        if ($d_deleted || (!defined $d_sfname) || ('.' eq $d_sfname)
             || ('..' eq $d_sfname) || !&is_valid_clusterno($d_clust) ) {
           next;
        }
        if ('d' ne $d_ftype) {
           $file_count++;
           next;
        }
        if (defined $lost_dirs{$d_clust}) {
           push( @result, $d_clust );
           if (1) {  # ($o_debugmode) {
              printf "Lost <DIR> entry found in dir cluster 0x0%X", $clust1;
              if ($clustno != $clust1) {
                 printf ",0x0%X", $clustno;
              }
              printf ", entry points to lost dir '%s' at cluster %d=0x0%X\n",
                     $d_sfname, $d_clust, $d_clust;
           }
        } else {
           print "ERROR: cross-linkage?  Non-lost <DIR> entry found in dir cluster 0x0%X", $clust1;
              if ($clustno != $clust1) {
                 printf ",0x0%X", $clustno;
              }
              printf ", entry points to dir '%s'", $d_sfname;
              printf " at cluster %d=0x0%X\n", $d_clust, $d_clust;
        }
     }#for
     $clustno = &Fat1Entry($clustno);  # next cluster in chain that makes up this directory
  }#while
  printf "Read1lostdir(0x0%X), lost folder contains %d files \& folders, including %d lost <DIR> entries.\n",
         $clust1, 1+$#result;
  return @result;
}# sub Read1lostdir

# Next, modify the %lost_dirs lookup table, so that each entry lists the other
# lost dirs that are subdirectories of it.  Also, create a table with the number of
# contained files & folders for each lost dir.
foreach $clust (@lost_folders) {
   @tmp = &Read1lostdir( $clust );
   if (-1 != $#tmp) {
      $lost_dirs{$clust} = join('|',@tmp);
   }
   $lost_file_counts{$clust} = $file_count;
}


# for debugging, generate displayable list of lost folder starting clusters w/ file counts
sub show_subdirlist {
   local( $inpl ) = @_;
   local( $result ) = '';
   local( $clus );
   local( @tmp ) = split( /\|/, $inpl );
   foreach $clus (@tmp) {
      $clus += 0;
      $result .= sprintf( ", %d=0x0%X", $clus, $clus );
      if (defined $lost_file_counts{$clus}) {
         $result .= ('[' . $lost_file_counts{$clus} .']');
      }
   }
   if ('' ne $result) {
      $result = substr( $result, 2 );
   }
   return $result;
}


# Next, we find the top recovered folder(s) by eliminating the referred-to folders
# from the list, for each recovered folder.
for ($i=0; $i <= $#lost_folders; ) {
   $clust = $lost_folders[$i];
   $subdirlist = $lost_dirs{$clust};
   printf "dbg: eliminating subdirs; %d lost folders, i=$i, clust=%d=0x0%X, subdirs='%s'\n",
          1+$#lost_folders, $clust, $clust, &show_subdirlist($subdirlist);
   if ($subdirlist ne '') {
      @tmp = split(/\|/,$subdirlist);
      foreach $subdir (@tmp) {
         $subdir += 0;  # convert string back to int
         if (defined $lost_dirs{$subdir}) {
            printf "  dbg: zap subdir clust %d=0x0%X\n", $subdir, $subdir;
            $lost_file_counts{$clust} += $lost_file_counts{$subdir};  # acredit the subdir's files to the parent
            delete $lost_dirs{$subdir};
            delete $lost_file_counts{$subdir};
         } else {
            printf "ERROR: cross-linkage?  cluster 0x0%X is subdir of more than one other!\n", $subdir;
         }
      }
      if (!defined $lost_dirs{$clust}) {
         printf "ERROR: I am my own grandpa!?!\n";
      }
      $lost_dirs{$clust} = '';
      $i = 0;
      @lost_folders = sort keys %lost_dirs;
   } else {
      $i++;
   }
}#for

print "Done eliminating lost subdirs.\n";
printf "%d highest-level lost folder(s) found.  These are their starting clusters:\n", 1+$#lost_folders;
print " " . &show_subdirlist(join('|',@lost_folders)) . "\n";



# result is a string representation of the cluster list passed as the input parameter(s).
# if $o_listfiles is <= 1 then result is exactly 16 characters long.
# if $o_listfiles is > 1 then result is at least 16 characters long.
sub show_hoplist {
   local( @hoplst ) = @_;
   local( $result ) = "";
   if (($o_listfiles <= 1) || ($#hoplst <= 1)) {
      if (0 == $#hoplst) {
         # just 1 cluster in cluster chain for this file
         $result = sprintf( "%07X", $hoplst[0] );
      } elsif (1 == $#hoplst) {
         # exactly 2 clusters in cluster chain for this file
         $result = sprintf( "%07X,%07X", $hoplst[0], $hoplst[1] );
      } else {
         # 3 or more clusters in cluster chain for this file; show the 1st & last
         $result = sprintf( "%07X~%07X", $hoplst[0], $hoplst[$#hoplst] );
      }
   } else {
      # there are 3 or more clusters, and we need to display them all
      local($i,$numconsec,$clust,$prev);
      $numconsec = 0;
      $clust = $hoplst[0];
      $result = sprintf("%07X", $hoplst[0]);
      for ($i=1; $i<=$#hoplst; $i++) {
         $prev = $clust;
         $clust = $hoplst[$i];
         if ($clust == (1+$prev)) {
            $numconsec++;
         } else {
            if ($numconsec) {
               $result .= sprintf("-%07X",$prev);
            }
            $result .= sprintf(",%07X",$clust);
         }
      }#for
      if ($numconsec) {
         $result .= sprintf("-%07X",$clust);
      }
   }
   local( $FAT_chain_ending ) = &Fat1Entry($hoplst[$#hoplst]);
   if ((0xFFFFFFF8 & $FAT_chain_ending) != 0x0FFFFFF8) {
      if (length($result) <= 8) {
         $result .= " ";
      }
      $result .= "?";  # bad cluster chain -- should end in -0x0FFFFFFF
   } elsif ($FAT_chain_ending != 0x0FFFFFFF) {
      print "dbg-ERR: HoplistTerm\n";
   }
   if (length($result) < 16) {
      $result .= " " x (16 - length($result));
   }
   if ($o_listfiles >= 3) {
      # show cluster list terminators (should be 0x0FFFFFFF)
      $result .= sprintf( " [%08X]", &Fat1Entry($hoplst[$#hoplst]) );
   }
   return $result;
} #show_hoplist


# read specified data cluster
sub fetchCluster {
   local( $clustnum ) = @_;
   if (!&is_valid_clusterno($clustnum)) {
      print "ERROR: bad cluster number $clustnum to fetchCluster()\n";
      &quit(2,1);
   }
   local($lba1) = &cluster_to_lba($clustnum);
   local($fofs) = $lba1 * 512;
   local($cluster) = &bigread1( $fofs, $b_sectorsPerCluster );
   if (($b_sectorsPerCluster * 512) != length($cluster)) {
      printf "ERROR: could not read cluster no. $clustnum (LBA %s = byte offset %s), only %d bytes were read.\n",
             &commas5($lba1), &commas5($fofs), length($cluster);
      &quit(2,1);
   }
   return $cluster;
} # sub fetchCluster


# input is a cluster, presumably part of a directory
# output is the number of non-null entries (including deleted file
# entries, longname fragments, etc.)  The last cluster in a directory
# usually has at least one null entry at the end.
sub count_DirEntriesInCluster {
   local( $cluster ) = @_;
   local( $result ) = 0;
   local( $i );
   for ($i=0; $i < length($cluster); $i+=32) {
      &process1dirent( $cluster, $i, $parent_inode );
      if ((defined $d_sfname) && ($d_ftype ne "\0")) {
         $result++;
      } else {
         last;
      }
   }
   return $result;
} # sub count_DirEntriesInCluster


$timer{'correct_FAT_from_directories'} = 0.0;

# the '-c' option ($o_correctFATfromDirs)
sub correct_FAT_from_directories {
   local( $inode, $fsize, $ftype, $clust, $lastc, $lastc_points_to, $i);
   $timer{'correct_FAT_from_directories'} -= (times)[0];
   print "enter correct_FAT_from_directories\n";
   local( @tofix_inode, @tofix_FSize, @tofix_nxtc );

   # First, make a list of files and folders which have bad cluster chains
   for ($inode=0; $inode <= $#in_Shortname; $inode++) {
      $clust = $in_FirstClust[$inode];
      $lastc = &LastClust($clust);  # sets @hoplist and $cluster_chain_missing as side-effect results
      if ($cluster_chain_missing >= 1) {
         # bad cluster chain!
         push( @tofix_inode, $inode );
         push( @tofix_FSize, $in_FSize[$inode] );
         push( @tofix_nxtc, $lastc );
      }
   }
   printf "%d inodes have bad cluster chains.\n", 1+$#tofix_inode;

   # Next, let's fix all the folders; we'll come back for the files
   for ($i=0; $i <= $#tofix_inode; $i++) {
      $inode = $tofix_inode[$i];
      $ftype = $in_FType[$inode];
      $lastc = $tofix_nxtc[$i];
      $lastc_points_to = &Fat1Entry($lastc);
      if ($o_debugmode) {
         printf "dbg/fix: $i -> $inode '$ftype'  @(0x0%X) = 0x0%X\n", $lastc, $lastc_points_to;
      }
      if ('d' eq $ftype) {
         if (($lastc_points_to != 0) && &is_valid_FAT_entry($lastc_points_to)) {
            printf "ERROR: inode $inode: can't correct FAT entry, it's already 0x0%X\n", $lastc_points_to;
         } else {
            if ($o_debugmode) {
               printf "inode $inode: correcting cluster 0x0%X from [%08X] to [0FFFFFFF].\n", $lastc, $lastc_points_to;
            }
            $cluster = &fetchCluster( $lastc );
            $num_dir_entries = &count_DirEntriesInCluster( $cluster );
            print "$num_dir_entries 32-byte entries found in cluster $lastc of folder at inode $inode.\n";
            if ($num_dir_entries < ($b_sectorsPerCluster * 512 / 32)) {  # 128 for 4KB clusters
               # This is the only cluster in the cluster chain for the folder
               # so this folder is completely repaired.
               $fat = &poke_longword_in_afat( 1, $lastc, 0x0FFFFFFF );
               if ($num_dir_entries > 0) {
                  printf "Folder '%s' (inode $inode) is completely repaired.\n", &inode2path($inode);
               } else {
                  printf "ERROR: Folder '%s' (inode $inode) doesn't look like a directory.\n", &inode2path($inode);
               }
            } else {
               printf "Warning: folder '%s' at inode $inode is probably incomplete.\n", &inode2path($inode);
               $fat = &poke_longword_in_afat( 1, $lastc, 0xFFFFFFFF );
               # terminate cluster chain invalidly, with 0xFFFFFFFF, so that
               # it'll still be flagged in the directory list.
            }
            splice( @tofix_inode, $i, 1 );
            splice( @tofix_nxtc, $i, 1 );
            splice( @tofix_FSize, $i, 1 );
            $i--;
         }
      }
   }#for

   # Now the files
   printf "dbg: files, pt1 (%d still need fixing)\n", 1+$#tofix_inode;
   for ($i=0; $i <= $#tofix_inode; $i++) {
      $inode = $tofix_inode[$i];
      $ftype = $in_FType[$inode];
      $lastc = $tofix_nxtc[$i];
      $lastc_points_to = &Fat1Entry($lastc);
      if (($lastc_points_to != 0) && &is_valid_FAT_entry($lastc_points_to)) {
         printf "ERROR: inode $inode: can't correct FAT entry, it's already 0x0%X\n", $lastc_points_to;
         splice( @tofix_inode, $i, 1 );
         splice( @tofix_nxtc, $i, 1 );
         splice( @tofix_FSize, $i, 1 );
         $i--;
      } elsif ('f' eq $ftype) {
         if ($o_debugmode) {
            printf "inode $inode: correcting cluster 0x0%X from [%08X] to [0FFFFFFF].\n", $lastc, $lastc_points_to;
         }
         $fat = &poke_longword_in_afat( 1, $lastc, 0x0FFFFFFF );
         if ($tofix_FSize[$i] <= (512 * $b_sectorsPerCluster)) {
            # it's only one cluster long, so we're done
            splice( @tofix_inode, $i, 1 );
            splice( @tofix_nxtc, $i, 1 );
            splice( @tofix_FSize, $i, 1 );
            $i--;
            printf "Success: fixed FAT for 1-cluster file, '%s' (inode $inode)\n", &inode2path($inode);
         } else {
            $tofix_nxtc[$i]++;
            $tofix_FSize[$i] -= (512 * $b_sectorsPerCluster);
         }
      } else {
         print "ERR: skipping inode $inode (%s)\n", &inode2path($inode);
      }
   }
   printf "dbg: files, pt2 (%d still need fixing)\n", 1+$#tofix_inode;
   while ($#tofix_inode >= 0) {
      printf "dbg: top-of-while (%d still need fixing)\n", 1+$#tofix_inode;
      for ($i=0; $i <= $#tofix_inode; $i++) {
         $inode = $tofix_inode[$i];
         $ftype = $in_FType[$inode];
         $lastc = $tofix_nxtc[$i];
         $lastc_points_to = &Fat1Entry($lastc);
         if (($lastc_points_to != 0) && &is_valid_FAT_entry($lastc_points_to)) {
            printf "ERROR: inode $inode: can't correct next FAT entry, it's already 0x0%X\n", $lastc_points_to;
            splice( @tofix_inode, $i, 1 );
            splice( @tofix_nxtc, $i, 1 );
            splice( @tofix_FSize, $i, 1 );
            $i--;
         } elsif ('f' eq $ftype) {
            if ($o_debugmode) {
               printf "inode $inode: correcting cluster 0x0%X from [0FFFFFFF] to [%08X].\n", $lastc-1, $lastc;
               printf "          and correcting cluster 0x0%X from [%08X] to [0FFFFFFF].\n", $lastc, $lastc_points_to;
            }
            $fat = &poke_longword_in_afat( 1, $lastc-1, $lastc );
            $fat = &poke_longword_in_afat( 1, $lastc, 0x0FFFFFFF );
            if ($tofix_FSize[$i] <= (512 * $b_sectorsPerCluster)) {
               # That's the last cluster, so we're done
               splice( @tofix_inode, $i, 1 );
               splice( @tofix_nxtc, $i, 1 );
               splice( @tofix_FSize, $i, 1 );
               $i--;
               printf "Success: fixed entire FAT chain for file, '%s' (inode $inode)\n", &inode2path($inode);
            } else {
               $tofix_nxtc[$i]++;
               $tofix_FSize[$i] -= (512 * $b_sectorsPerCluster);
            }
         } else {
            print "ERR: skipping inode $inode (%s)\n", &inode2path($inode);
         }
      }#for
   }#while
   printf "dbg: files, pt2 done (%d still need fixing)\n", 1+$#tofix_inode;

   &build_LastClust2inode;
   $timer{'correct_FAT_from_directories'} += (times)[0];
   printf "exit from correct_FAT_from_directories (%.2f sec runtime)\n", $timer{'correct_FAT_from_directories'};
} # sub correct_FAT_from_directories


if ($o_correctFATfromDirs) {
   &correct_FAT_from_directories;
}


# display full inode/file list if '-l' was specified:
$fsize = 0;
$max_possible_size = $min_possible_size = 0;
$num_size_errors = 0;
if ($o_listfiles) {
   printf "dbg: list all %s files and folders:\n", &commas5(1+$#in_Shortname);
   for ($inode=$nxt_inode_to_examine; $inode <= $#in_Shortname; $inode++) {
      $bad_start_clust = 0;
      $ftype = $in_FType[$inode];
      $clust = $in_FirstClust[$inode];
      $path = &inode2path($inode);
      if (! &is_valid_clusterno($clust)) {
         $lastc = 0;
         $bad_start_clust = 1;
      } else {
         $lastc = &LastClust($clust);  # sets @hoplist as a side-effect result
         $inode2 = $LastClust2inode{$lastc};
         if ($inode2 != $inode) {
            printf "ERR: inode=%d -> clust=%s -> lastclust=%s -> inode=%d:\n",
                   $inode, &DispFatEnt($clust), &DispFatEnt($lastc), $inode2;
         }
      }
      printf "%5d:", $inode;
      if (!$o_minmem) {
         $fsize = $in_FSize[$inode];
         if ((0 == $fsize) && (0 == $clust) && ('d' ne $ftype)) {
            $bad_start_clust = 0;  # 0-length file doesn't need any clusters
         }
      }
      if ($o_minmem) {
         if ('d' eq $ftype) {
            print " <DIR> ";
         } else {
            # in '-m' mode we don't store the file sizes, so we can't print 'em
            print "       ";
         }
      } else {
         if ('d' eq $ftype) {
            print "   <DIR>   ";
            # print "[dbg: siz=$fsize] ";  # seems to always be zero
         } else {
            printf "%11d", $fsize;
         }
      }
      if ($bad_start_clust) {
         printf "  %07X=BAD     ", $clust;
      } elsif (0 == $clust) {
         print "  0000000         ";
      } else {
         if (($clust != $hoplist[0]) || ($lastc != $hoplist[$#hoplist])) {
            printf " {ERR: hoplist,  F=0x0%X, h[0]=0x0%X;  L=0x0%X, h[%d]=0x0%X} ",
                   $clust, $hoplist[0], $lastc, $#hoplist, $hoplist[$#hoplist];
         }

         # sanity check: is the length of the cluster chain consistent with the file size in the directory entry?

         $max_possible_size = ($b_bytesPerSector * $b_sectorsPerCluster) * (1 + $#hoplist);
         $min_possible_size = 1 + (($b_bytesPerSector * $b_sectorsPerCluster) * $#hoplist);
         if (('d' ne $ftype) && (($fsize > $max_possible_size) || ($fsize < $min_possible_size))) {
            print "* ";  # size error!
            $num_size_errors++;
         } elsif (((1 == $cluster_chain_missing) && ($fsize != 0)) || ($cluster_chain_missing > 1)) {
            print "* ";  # bad cluster chain!
            $num_size_errors++;
         } else {
            print "  ";
         }

         print &show_hoplist( @hoplist );  # 16 (or more) characters long
         # if ($lastc == $clust) {
         #    # just 1 cluster in cluster chain for this file
         #    printf "  %07X         ", $clust;
         # } elsif (&FatEntry($clust) == $lastc) {
         #    # exactly 2 clusters in cluster chain for this file
         #    printf "  %07X, %07X", $clust, $lastc;
         # } else {
         #    # 3 or more clusters in cluster chain for this file
         #    printf "  %07X..%07X", $clust, $lastc;
         # }
        #  printf " (%d..%d) ", &cluster_to_lba($clust), &cluster_to_lba($lastc)+$b_sectorsPerCluster-1;
      }
      print " $path\n";
      if ($o_longnames && (defined $in_Longname[$inode]) && ('' ne $in_Longname[$inode])) {
         print '' . (' ' x 36) . '"' . $in_Longname[$inode] . '"' . "\n";
      }
   }#for
}#if
if ($num_size_errors) {
   print "Warning!  $num_size_errors files have inconsistent sizes (cluster chain length inconsistent with file size in directory entry).\n";
}


if ($o_reconcilewrite) {
   if (($fat1fixes > 0) || ($o_correctFATfromDirs)) {
      print "Writing corrected FAT #1 to disk.\n";
      &write_FAT1;
      print "Done writing.\n";
   } else {
      print "It was not necessary to write FAT #1 to disk, because no sectors in FAT #1 were corrected.\n";
   }
}


# Look up each of the sector numbers specified on the command line:
if ($o_lookup_sector) {
   print "Look up file names from sectors:\n";
   for ($i=2; $i<=$#ARGV; $i++) {
      $sect = &hex_or_positive_decimal( $ARGV[$i] );
      $clust = &lba_to_cluster( $sect );
      if ($clust < 2) {
         print "sector $sect is in the system areas.\n";
      } else {
         $nxt = &Fat1Entry( $clust );
         if (0 == $nxt) {
            print "sector $sect is in free space.\n";
         } elsif (0x0FFFFFF7 == $nxt) {
            printf "sector $sect is in cluster %d=0x0%X, which is marked BAD in the FAT.\n", $clust, $clust;
         } elsif (-1 == $nxt) {
            print "Could not deduce what cluster $sect is in.\n";
         } else {
            $last = &LastClust( $clust );
            $inode = $LastClust2inode{$last};
            if (!defined $inode) {
               printf "sector $sect is in cluster %d=0x0%X, which has FAT entry %s and is part of a lost cluster chain ending at cluster %d=0x0%X.\n", $clust, $clust, &DispFatEnt($nxt), $last, $last;
            } else {
               printf "sector $sect is in %s\n", &inode2path($inode);
            }
         }
      }
   }
}

&quit(0);

__END__

